/*
***********************************************************************
* COPYRIGHT AND WARRANTY INFORMATION
*
* Copyright 2001, International Telecommunications Union, Geneva
*
* DISCLAIMER OF WARRANTY
*
* These software programs are available to the user without any
* license fee or royalty on an "as is" basis. The ITU disclaims
* any and all warranties, whether express, implied, or
* statutory, including any implied warranties of merchantability
* or of fitness for a particular purpose.  In no event shall the
* contributor or the ITU be liable for any incidental, punitive, or
* consequential damages of any kind whatsoever arising from the
* use of these programs.
*
* This disclaimer of warranty extends to the user of these programs
* and user's customers, employees, agents, transferees, successors,
* and assigns.
*
* The ITU does not represent or warrant that the programs furnished
* hereunder are free of infringement of any third-party patents.
* Commercial implementations of ITU-T Recommendations, including
* shareware, may be subject to royalty fees to patent holders.
* Information regarding the ITU-T patent policy is available from
* the ITU Web site at http://www.itu.int.
*
* THIS IS NOT A GRANT OF PATENT RIGHTS - SEE THE ITU-T PATENT POLICY.
************************************************************************
*/

/*!
 ***********************************************************************
 * \file macroblock.c
 *
 * \brief
 *     Decode a Macroblock
 *
 * \author
 *    Main contributors (see contributors.h for copyright, address and affiliation details)
 *    - Inge Lille-Langy               <inge.lille-langoy@telenor.com>
 *    - Rickard Sjoberg                 <rickard.sjoberg@era.ericsson.se>
 *    - Jani Lainema                    <jani.lainema@nokia.com>
 *    - Sebastian Purreiter             <sebastian.purreiter@mch.siemens.de>
 *    - Thomas Wedi                     <wedi@tnt.uni-hannover.de>
 *    - Detlev Marpe                    <marpe@hhi.de>
 *    - Gabi Blaettermann               <blaetter@hhi.de>
 *    - Ye-Kui Wang                     <wyk@ieee.org>
 ***********************************************************************
*/

#include "contributors.h"

#include <math.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>

#include "global.h"
#include "mbuffer.h"
#include "elements.h"
#include "errorconcealment.h"
#include "macroblock.h"
#include "fmo.h"
#include "vlc.h"
#include "image.h"

#if TRACE
#define TRACE_STRING(s) strncpy(currSE.tracestring, s, TRACESTRING_SIZE)
#else
#define TRACE_STRING(s) // do nothing
#endif

extern StorablePicture* dec_picture;

static void SetMotionVectorPredictor (struct img_par  *img,
                                      int             *pmv_x,
                                      int             *pmv_y,
                                      int             ref_frame,
                                      int             list,
                                      int             ***refPic,
                                      int             ****tmp_mv,
                                      int             block_x,
                                      int             block_y,
                                      int             blockshape_x,
                                      int             blockshape_y);


/*!
 ************************************************************************
 * \brief
 *   maps curr_mb_nr to the linear macroblock adress in img->map_mb_nr
 ************************************************************************
 */
void setMapMB_nr (struct img_par *img)
{
  int max_mb_row, x;
  
  if (img->mb_frame_field_flag)
  {
    max_mb_row = (img->width/MB_BLOCK_SIZE);
    x = img->current_mb_nr/(2*max_mb_row);
    if (img->current_mb_nr % 2) 
      img->map_mb_nr = img->current_mb_nr/2 + (x+1) * max_mb_row;
    else
      img->map_mb_nr = img->current_mb_nr/2 + x * max_mb_row;
  }
  else
    img->map_mb_nr = img->current_mb_nr;
}


/*!
 ************************************************************************
 * \brief
 *    returns 1 if the macroblock at the given address is available
 ************************************************************************
 */
int mb_is_available(int mbAddr, int currMbAddr)
{
  if ((mbAddr < 0) || (mbAddr > ((int)img->PicSizeInMbs - 1)))
    return 0;

  // the following line checks both: slice number and if the mb has been decoded
  if (img->mb_data[mbAddr].slice_nr != img->mb_data[currMbAddr].slice_nr)
    return 0;
  
  return 1;
}


/*!
 ************************************************************************
 * \brief
 *    Checks the availability of neighboring macroblocks of
 *    the current macroblock for prediction and context determination;
 ************************************************************************
 */
void CheckAvailabilityOfNeighbors(struct img_par *img)
{
  int i,j;
  const int mb_width = img->width/MB_BLOCK_SIZE;
  const int mb_nr = img->current_mb_nr;
  Macroblock *currMB = &img->mb_data[mb_nr];
  int check_value;

  // mark all neighbors as unavailable
  for (i=0; i<3; i++)
    for (j=0; j<3; j++)
      currMB->mb_available[i][j]=NULL;

  currMB->mb_available[1][1]=currMB; // current MB

  if (img->mb_frame_field_flag)
  {
    currMB->mbAddrA = 2 * (mb_nr/2 - 1);
    currMB->mbAddrB = 2 * (mb_nr/2 - img->PicWidthInMbs);
    currMB->mbAddrC = 2 * (mb_nr/2 - img->PicWidthInMbs + 1);
    currMB->mbAddrD = 2 * (mb_nr/2 - img->PicWidthInMbs - 1);
    
    currMB->mbAvailA = mb_is_available(currMB->mbAddrA, mb_nr) && ((mb_nr/2 % img->PicWidthInMbs)!=0);
    currMB->mbAvailB = mb_is_available(currMB->mbAddrB, mb_nr);
    currMB->mbAvailC = mb_is_available(currMB->mbAddrC, mb_nr) && (((mb_nr/2 +1) % img->PicWidthInMbs)!=0);
    currMB->mbAvailD = mb_is_available(currMB->mbAddrD, mb_nr) && ((mb_nr/2 % img->PicWidthInMbs)!=0);
  }
  else
  {
    currMB->mbAddrA = mb_nr - 1;
    currMB->mbAddrB = mb_nr - img->PicWidthInMbs;
    currMB->mbAddrC = mb_nr - img->PicWidthInMbs + 1;
    currMB->mbAddrD = mb_nr - img->PicWidthInMbs - 1;

    currMB->mbAvailA = mb_is_available(currMB->mbAddrA, mb_nr) && ((mb_nr % img->PicWidthInMbs)!=0);
    currMB->mbAvailB = mb_is_available(currMB->mbAddrB, mb_nr);
    currMB->mbAvailC = mb_is_available(currMB->mbAddrC, mb_nr) && (((mb_nr+1) % img->PicWidthInMbs)!=0);
    currMB->mbAvailD = mb_is_available(currMB->mbAddrD, mb_nr) && ((mb_nr % img->PicWidthInMbs)!=0);
  }

  if (currMB->mbAvailA)
    currMB->mb_available[1][0]=&(img->mb_data[currMB->mbAddrA]);
  if (currMB->mbAvailB)
    currMB->mb_available[0][1]=&(img->mb_data[currMB->mbAddrC]);
  if(currMB->mbAvailC)
    currMB->mb_available[0][2]=&(img->mb_data[currMB->mbAddrC]);
  if (currMB->mbAvailD)
    currMB->mb_available[0][0]=&(img->mb_data[currMB->mbAddrD]);

}


void CheckAvailabilityOfNeighborsForAff(struct img_par *img)
{
  const int mb_width = img->width/MB_BLOCK_SIZE;
  const int mb_nr = img->current_mb_nr;

  Macroblock *currMB = &img->mb_data[mb_nr];
  // Check Field to the left
  if ((img->pix_x >= MB_BLOCK_SIZE) && (currMB->slice_nr == img->mb_data[mb_nr-1].slice_nr))
    currMB->field_available[1]=&(img->mb_data[mb_nr-1]);
  else
    currMB->field_available[1]=NULL;
  
  // Check Field above
  if ((img->pix_y >= 2*MB_BLOCK_SIZE) && (currMB->slice_nr == img->mb_data[mb_nr-mb_width*2].slice_nr))
    currMB->field_available[0]=&(img->mb_data[mb_nr-mb_width*2]);
  else
    currMB->field_available[0]=NULL;

}

void get_mb_pos (int mb_addr, int *x, int*y)
{

  if (img->mb_frame_field_flag)
  {
    *x = ((mb_addr/2) % img->PicWidthInMbs) * MB_BLOCK_SIZE;
    *y = ( ((mb_addr/2) / img->PicWidthInMbs)  * 2 + (mb_addr%2)) * MB_BLOCK_SIZE;
  }
  else
  {
    *x = (mb_addr % img->PicWidthInMbs) * MB_BLOCK_SIZE;
    *y = (mb_addr / img->PicWidthInMbs) * MB_BLOCK_SIZE;
  }
}

/*!
 ************************************************************************
 * \brief
 *    get neighbouring positions for non-aff coding
 * \param curr_mb_nr
 *   current macroblock number (decoding order)
 * \param xN
 *    input x position
 * \param yN
 *    input y position
 * \param luma
 *    1 if luma coding, 0 for chroma
 * \param MbAddr
 *    output address of neighbouring macroblock
 * \param MbAvail
 *    output availability of neighbour
 * \param xW
 *    output neighbouring x position
 * \param yW
 *    output neighbouring y position
 ************************************************************************
 */
void getNonAffNeighbour(int curr_mb_nr, int xN, int yN, int luma, PixelPos *pix)
{
  Macroblock *currMb = &img->mb_data[curr_mb_nr];
  int maxWH;

  if (luma)
    maxWH = 16;
  else
    maxWH = 8;

  if ((xN<0)&&(yN<0))
  {
    pix->mb_addr   = currMb->mbAddrD;
    pix->available = currMb->mbAvailD;
  }
  else
  if ((xN<0)&&((yN>=0)&&(yN<maxWH)))
  {
    pix->mb_addr  = currMb->mbAddrA;
    pix->available = currMb->mbAvailA;
  }
  else
  if (((xN>=0)&&(xN<maxWH))&&(yN<0))
  {
    pix->mb_addr  = currMb->mbAddrB;
    pix->available = currMb->mbAvailB;
  }
  else
  if (((xN>=0)&&(xN<maxWH))&&((yN>=0)&&(yN<maxWH)))
  {
    pix->mb_addr  = curr_mb_nr;
    pix->available = 1;
  }
  else
  if ((xN>=maxWH)&&(yN<0))
  {
    pix->mb_addr  = currMb->mbAddrC;
    pix->available = currMb->mbAvailC;
  }
  else 
  {
    pix->available = 0;
  }
  if (pix->available)
  {
    pix->x = (xN + maxWH) % maxWH;
    pix->y = (yN + maxWH) % maxWH;
    get_mb_pos(pix->mb_addr, &(pix->pos_x), &(pix->pos_y));
    if (luma)
    {
      pix->pos_x += pix->x;
      pix->pos_y += pix->y;
    }
    else
    {
      pix->pos_x = (pix->pos_x/2) + pix->x;
      pix->pos_y = (pix->pos_y/2) + pix->y;
    }
  }
}

/*!
 ************************************************************************
 * \brief
 *    get neighbouring positions for aff coding
 * \param curr_mb_nr
 *   current macroblock number (decoding order)
 * \param xN
 *    input x position
 * \param yN
 *    input y position
 * \param luma
 *    1 if luma coding, 0 for chroma
 * \param MbAddr
 *    output address of neighbouring macroblock
 * \param MbAvail
 *    output availability of neighbour
 * \param xW
 *    output neighbouring x position
 * \param yW
 *    output neighbouring y position
 ************************************************************************
 */
void getAffNeighbour(int curr_mb_nr, int xN, int yN, int luma, PixelPos *pix)
{
  Macroblock *currMb = &img->mb_data[curr_mb_nr];
  int maxWH;
  int yM;

  if (luma)
    maxWH = 16;
  else
    maxWH = 8;

  // initialize to "not available"
  pix->available = 0;

  if(yN > (maxWH - 1))
  {
    return;
  }
  if ((xN > (maxWH -1)) && ((yN >= 0)&&(yN < (maxWH ))))
  {
    return;
  }

  if (xN < 0)
  {
    if (yN < 0)
    {
      if(!currMb->mb_field)
      {
        // frame
        if (curr_mb_nr%2 == 0)
        {
          // top
          pix->mb_addr  = currMb->mbAddrD  + 1;
          pix->available = currMb->mbAvailD;
           yM      = yN;
        }
        else
        {
          // bottom
          pix->mb_addr  = currMb->mbAddrA;
          pix->available = currMb->mbAvailA;
          if (currMb->mbAvailA)
          {
            if(!img->mb_data[currMb->mbAddrA].mb_field)
            {
               yM = yN;
            }
            else
            {
              (pix->mb_addr)++;
               yM = (yN + maxWH) >> 1;
            }
          }
        }
      }
      else
      {
        // field
        if(curr_mb_nr % 2 == 0)
        {
          // top
          pix->mb_addr  = currMb->mbAddrD;
          pix->available = currMb->mbAvailD;
          if (currMb->mbAvailD)
          {
            if(!img->mb_data[currMb->mbAddrD].mb_field)
            {
              (pix->mb_addr)++;
               yM = 2 * yN;
            }
            else
            {
               yM = yN;
            }
          }
        }
        else
        {
          // bottom
          pix->mb_addr  = currMb->mbAddrD+1;
          pix->available = currMb->mbAvailD;
           yM      = yN;
        }
      }
    }
    else
    { // xN < 0 && yN >= 0
      if ((yN >= 0) && (yN <maxWH))
      {
        if (!currMb->mb_field)
        {
          // frame
          if(curr_mb_nr % 2 == 0)
          {
            // top
            pix->mb_addr  = currMb->mbAddrA;
            pix->available = currMb->mbAvailA;
            if (currMb->mbAvailA)
            {
              if(!img->mb_data[currMb->mbAddrA].mb_field)
              {
                 yM = yN;
              }
              else
              {
                if (yN %2 == 0)
                {
                   yM = yN>> 1;
                }
                else
                {
                  (pix->mb_addr)++;
                   yM = yN>> 1;
                }
              }
            }
          }
          else
          {
            // bottom
            pix->mb_addr  = currMb->mbAddrA;
            pix->available = currMb->mbAvailA;
            if (currMb->mbAvailA)
            {
              if(!img->mb_data[currMb->mbAddrA].mb_field)
              {
                (pix->mb_addr)++;
                 yM = yN;
              }
              else
              {
                if (yN %2 == 0)
                {
                   yM = (yN + maxWH) >> 1;
                }
                else
                {
                  (pix->mb_addr)++;
                   yM = (yN + maxWH) >> 1;
                }
              }
            }
          }
        }
        else
        {
          // field
          if (curr_mb_nr % 2 == 0)
          {
            // top
            pix->mb_addr  = currMb->mbAddrA;
            pix->available = currMb->mbAvailA;
            if (currMb->mbAvailA)
            {
              if(!img->mb_data[currMb->mbAddrA].mb_field)
              {
                if (yN < (maxWH / 2))
                {
                   yM = yN << 1;
                }
                else
                {
                  (pix->mb_addr)++;
                   yM = (yN << 1 ) - maxWH;
                }
              }
              else
              {
                 yM = yN;
              }
            }
          }
          else
          {
            // bottom
            pix->mb_addr  = currMb->mbAddrA;
            pix->available = currMb->mbAvailA;
            if (currMb->mbAvailA)
            {
              if(!img->mb_data[currMb->mbAddrA].mb_field)
              {
                if (yN < (maxWH / 2))
                {
                   yM = (yN << 1) + 1;
                }
                else
                {
                  (pix->mb_addr)++;
                   yM = (yN << 1 ) + 1 - maxWH;
                }
              }
              else
              {
                 (pix->mb_addr)++;
                 yM = yN;
              }
            }
          }
        }
      }
    }
  }
  else
  {
    // xN >= 0
    if ((xN >= 0)&&(xN <maxWH))
    {
      if (yN<0)
      {
        if (!currMb->mb_field)
        {
          //frame
          if (curr_mb_nr % 2 == 0)
          {
            //top
            pix->mb_addr  = currMb->mbAddrB + 1;
            pix->available = currMb->mbAvailB;
             yM      = yN;
          }
          else
          {
            // bottom
            pix->mb_addr  = curr_mb_nr - 1;
            pix->available = 1;
             yM      = yN;
          }
        }
        else
        {
          // field
          if (curr_mb_nr % 2 == 0)
          {
            // top
            pix->mb_addr  = currMb->mbAddrB;
            pix->available = currMb->mbAvailB;
            if (currMb->mbAvailB)
            {
              if(!img->mb_data[currMb->mbAddrB].mb_field)
              {
                (pix->mb_addr)++;
                 yM = 2* yN;
              }
              else
              {
                 yM = yN;
              }
            }
          }
          else
          {
            // bottom
            pix->mb_addr  = currMb->mbAddrB + 1;
            pix->available = currMb->mbAvailB;
             yM      = yN;
          }
        }
      }
      else
      {
        // yN >=0
        if ((yN >= 0) && (yN <maxWH))
        {
          pix->mb_addr  = curr_mb_nr;
          pix->available = 1;
           yM      = yN;
        }
      }
    }
    else
    {
      // xN >= maxWH
      if(yN < 0)
      {
        if (!currMb->mb_field)
        {
          // frame
          if (curr_mb_nr % 2 == 0)
          {
            // top
            pix->mb_addr  = currMb->mbAddrC + 1;
            pix->available = currMb->mbAvailC;
             yM      = yN;
          }
          else
          {
            // bottom
            pix->available = 0;
          }
        }
        else
        {
          // field
          if (curr_mb_nr % 2 == 0)
          {
            // top
            pix->mb_addr  = currMb->mbAddrC;
            pix->available = currMb->mbAvailC;
            if (currMb->mbAvailC)
            {
              if(!img->mb_data[currMb->mbAddrC].mb_field)
              {
                (pix->mb_addr)++;
                 yM = 2* yN;
              }
              else
              {
                 yM = yN;
              }
            }
          }
          else
          {
            // bottom
            pix->mb_addr  = currMb->mbAddrC + 1;
            pix->available = currMb->mbAvailC;
             yM      = yN;
          }
        }
      }
    }
  }
  if (pix->available)
  {
    pix->x = (xN + maxWH) % maxWH;
    pix->y = (yM + maxWH) % maxWH;
    get_mb_pos(pix->mb_addr, &(pix->pos_x), &(pix->pos_y));
    if (luma)
    {
      pix->pos_x += pix->x;
      pix->pos_y += pix->y;
    }
    else
    {
      pix->pos_x = (pix->pos_x/2) + pix->x;
      pix->pos_y = (pix->pos_y/2) + pix->y;
    }
  }
}

void getNeighbour(int curr_mb_nr, int xN, int yN, int luma, PixelPos *pix)
{
  if (img->mb_frame_field_flag)
    getAffNeighbour(curr_mb_nr, xN, yN, luma, pix);
  else
    getNonAffNeighbour(curr_mb_nr, xN, yN, luma, pix);
}


void getLuma4x4Neighbour (int curr_mb_nr, int block_x, int block_y, int rel_x, int rel_y, PixelPos *pix)
{
  int x = 4* block_x + rel_x;
  int y = 4* block_y + rel_y;

  getNeighbour(curr_mb_nr, x, y, 1, pix);

  if (pix->available)
  {
    pix->x /= 4;
    pix->y /= 4;
    pix->pos_x /= 4;
    pix->pos_y /= 4;
  }
}

void getChroma4x4Neighbour (int curr_mb_nr, int block_x, int block_y, int rel_x, int rel_y, PixelPos *pix)
{
  int x = 4* block_x + rel_x;
  int y = 4* block_y + rel_y;

  getNeighbour(curr_mb_nr, x, y, 0, pix);

  if (pix->available)
  {
    pix->x /= 4;
    pix->y /= 4;
    pix->pos_x /= 4;
    pix->pos_y /= 4;
  }
}

/*!
 ************************************************************************
 * \brief
 *    initializes the current macroblock
 ************************************************************************
 */
void start_macroblock(struct img_par *img,struct inp_par *inp, int CurrentMBInScanOrder)
{
  int i,j,k,l;
  Macroblock *currMB;   // intialization code deleted, see below, StW

  assert (img->current_mb_nr >=0 && img->current_mb_nr < img->PicSizeInMbs);

  currMB = &img->mb_data[img->current_mb_nr];

  /* Update coordinates of the current macroblock */
  if (img->mb_frame_field_flag)
  {
    img->mb_x = (img->current_mb_nr)%((2*img->width)/MB_BLOCK_SIZE);
    img->mb_y = 2*((img->current_mb_nr)/((2*img->width)/MB_BLOCK_SIZE));

    if (img->mb_x % 2)
    {
      img->mb_y++;
    }

    img->mb_x /= 2;
  }
  else
  {
    img->mb_x = (img->current_mb_nr)%(img->width/MB_BLOCK_SIZE);
    img->mb_y = (img->current_mb_nr)/(img->width/MB_BLOCK_SIZE);
  }
  
  /* Define vertical positions */
  img->block_y = img->mb_y * BLOCK_SIZE;      /* luma block position */
  img->pix_y   = img->mb_y * MB_BLOCK_SIZE;   /* luma macroblock position */
  img->pix_c_y = img->mb_y * MB_BLOCK_SIZE/2; /* chroma macroblock position */
  
  /* Define horizontal positions */
  img->block_x = img->mb_x * BLOCK_SIZE;      /* luma block position */
  img->pix_x   = img->mb_x * MB_BLOCK_SIZE;   /* luma pixel position */
  img->pix_c_x = img->mb_x * MB_BLOCK_SIZE/2; /* chroma pixel position */

  // Save the slice number of this macroblock. When the macroblock below
  // is coded it will use this to decide if prediction for above is possible
  currMB->slice_nr = img->current_slice_nr;
  
  CheckAvailabilityOfNeighbors(img);

  // Reset syntax element entries in MB struct
  currMB->qp          = img->qp ;
  currMB->mb_type     = 0;
  currMB->delta_quant = 0;
  currMB->cbp         = 0;
  currMB->cbp_blk     = 0;
  currMB->c_ipred_mode= DC_PRED_8; //GB

  for (l=0; l < 2; l++)
    for (j=0; j < BLOCK_MULTIPLE; j++)
      for (i=0; i < BLOCK_MULTIPLE; i++)
        for (k=0; k < 2; k++)
          currMB->mvd[l][j][i][k] = 0;

  currMB->cbp_bits   = 0;

  // initialize img->m7 for ABT
  for (j=0; j<MB_BLOCK_SIZE; j++)
    for (i=0; i<MB_BLOCK_SIZE; i++)
      img->m7[i][j] = 0;

  // store filtering parameters for this MB 
  currMB->lf_disable = img->currentSlice->LFDisableIdc;
  currMB->lf_alpha_c0_offset = img->currentSlice->LFAlphaC0Offset;
  currMB->lf_beta_offset = img->currentSlice->LFBetaOffset;

}

/*!
 ************************************************************************
 * \brief
 *    set coordinates of the next macroblock
 *    check end_of_slice condition 
 ************************************************************************
 */
int exit_macroblock(struct img_par *img,struct inp_par *inp,int eos_bit)
{
  Slice *currSlice = img->currentSlice;

  //! The if() statement below resembles the original code, which tested 
  //! img->current_mb_nr == img->PicSizeInMbs.  Both is, of course, nonsense
  //! In an error prone environment, one can only be sure to have a new
  //! picture by checking the tr of the next slice header!

// printf ("exit_macroblock: FmoGetLastMBOfPicture %d, img->current_mb_nr %d\n", FmoGetLastMBOfPicture(), img->current_mb_nr);
  img->num_dec_mb++;
  
  if (img->num_dec_mb == img->PicSizeInMbs)
//  if (img->current_mb_nr == FmoGetLastMBOfPicture(currSlice->structure))
  {
    if (currSlice->next_header != EOS)
      currSlice->next_header = SOP;
    return TRUE;
  }
  // ask for last mb in the slice  UVLC
  else
  {
// printf ("exit_macroblock: Slice %d old MB %d, now using MB %d\n", img->current_slice_nr, img->current_mb_nr, FmoGetNextMBNr (img->current_mb_nr));

    img->current_mb_nr = FmoGetNextMBNr (img->current_mb_nr);

    if (img->current_mb_nr == -1)     // End of Slice group, MUST be end of slice
    {
      assert (nal_startcode_follows (img, inp, eos_bit) == TRUE);
      return TRUE;
    }

    if(nal_startcode_follows(img, inp, eos_bit) == FALSE) 
      return FALSE;

    if(img->type == I_SLICE  || img->type == SI_SLICE || active_pps->entropy_coding_mode == CABAC)
      return TRUE;
    if(img->cod_counter<=0)
      return TRUE;
    return FALSE;
  }
}

/*!
 ************************************************************************
 * \brief
 *    Interpret the mb mode for P-Frames
 ************************************************************************
 */
void interpret_mb_mode_P(struct img_par *img)
{
  int i;
  const int ICBPTAB[6] = {0,16,32,15,31,47};
  Macroblock *currMB = &img->mb_data[img->current_mb_nr];
  int         mbmode = currMB->mb_type;

#define ZERO_P8x8     (mbmode==5)
#define MODE_IS_P8x8  (mbmode==4 || mbmode==5)
#define MODE_IS_I4x4  (mbmode==6)
#define I16OFFSET     (mbmode-7)

  if(mbmode <4)
  {
    currMB->mb_type = mbmode;
    for (i=0;i<4;i++)
    {
      currMB->b8mode[i]   = mbmode;
      currMB->b8pdir[i]   = 0;
    }
  }
  else if(MODE_IS_P8x8)
  {
    currMB->mb_type = P8x8;
    img->allrefzero = ZERO_P8x8;
  }
  else if(MODE_IS_I4x4)
  {
    currMB->mb_type = I4MB;
    for (i=0;i<4;i++)
    {
      currMB->b8mode[i] = IBLOCK;
      currMB->b8pdir[i] = -1;
    }
  }
  else
  {
    currMB->mb_type = I16MB;
    for (i=0;i<4;i++) {currMB->b8mode[i]=0; currMB->b8pdir[i]=-1; }
    currMB->cbp= ICBPTAB[(I16OFFSET)>>2];
    currMB->i16mode = (I16OFFSET) & 0x03;
  }
}

/*!
 ************************************************************************
 * \brief
 *    Interpret the mb mode for I-Frames
 ************************************************************************
 */
void interpret_mb_mode_I(struct img_par *img)
{
  int i;
  const int ICBPTAB[6] = {0,16,32,15,31,47};
  Macroblock *currMB   = &img->mb_data[img->current_mb_nr];
  int         mbmode   = currMB->mb_type;

  if (mbmode==0)
  {
    currMB->mb_type = I4MB;
    for (i=0;i<4;i++) {currMB->b8mode[i]=IBLOCK; currMB->b8pdir[i]=-1; }
  }
  else
  {
    currMB->mb_type = I16MB;
    for (i=0;i<4;i++) {currMB->b8mode[i]=0; currMB->b8pdir[i]=-1; }
    currMB->cbp= ICBPTAB[(mbmode-1)>>2];
    currMB->i16mode = (mbmode-1) & 0x03;
  }
}

/*!
 ************************************************************************
 * \brief
 *    Interpret the mb mode for B-Frames
 ************************************************************************
 */
void interpret_mb_mode_B(struct img_par *img)
{
  static const int offset2pdir16x16[12]   = {0, 0, 1, 2, 0,0,0,0,0,0,0,0};
  static const int offset2pdir16x8[22][2] = {{0,0},{0,0},{0,0},{0,0},{0,0},{0,0},{1,1},{0,0},{0,1},{0,0},{1,0},
                                             {0,0},{0,2},{0,0},{1,2},{0,0},{2,0},{0,0},{2,1},{0,0},{2,2},{0,0}};
  static const int offset2pdir8x16[22][2] = {{0,0},{0,0},{0,0},{0,0},{0,0},{0,0},{0,0},{1,1},{0,0},{0,1},{0,0},
                                             {1,0},{0,0},{0,2},{0,0},{1,2},{0,0},{2,0},{0,0},{2,1},{0,0},{2,2}};

  const int ICBPTAB[6] = {0,16,32,15,31,47};
  Macroblock *currMB = &img->mb_data[img->current_mb_nr];

  int i, mbmode;
  int mbtype  = currMB->mb_type;
  int *b8mode = currMB->b8mode;
  int *b8pdir = currMB->b8pdir;

  //--- set mbtype, b8type, and b8pdir ---
  if (mbtype==0)       // direct
  {
      mbmode=0;       for(i=0;i<4;i++) {b8mode[i]=0;          b8pdir[i]=2; }
  }
  else if (mbtype==23) // intra4x4
  {
    mbmode=I4MB;    for(i=0;i<4;i++) {b8mode[i]=IBLOCK;     b8pdir[i]=-1; }
  }
  else if (mbtype>23) // intra16x16
  {
    mbmode=I16MB;   for(i=0;i<4;i++) {b8mode[i]=0;          b8pdir[i]=-1; }
    currMB->cbp     = ICBPTAB[(mbtype-24)>>2];
    currMB->i16mode = (mbtype-24) & 0x03;
  }
  else if (mbtype==22) // 8x8(+split)
  {
    mbmode=P8x8;       // b8mode and pdir is transmitted in additional codewords
  }
  else if (mbtype<4)   // 16x16
  {
    mbmode=1;       for(i=0;i<4;i++) {b8mode[i]=1;          b8pdir[i]=offset2pdir16x16[mbtype]; }
  }
  else if (mbtype%2==0) // 16x8
  {
    mbmode=2;       for(i=0;i<4;i++) {b8mode[i]=2;          b8pdir[i]=offset2pdir16x8 [mbtype][i/2]; }
  }
  else
  {
    mbmode=3;       for(i=0;i<4;i++) {b8mode[i]=3;          b8pdir[i]=offset2pdir8x16 [mbtype][i%2]; }
  }
  currMB->mb_type = mbmode;
}
/*!
 ************************************************************************
 * \brief
 *    Interpret the mb mode for SI-Frames
 ************************************************************************
 */
void interpret_mb_mode_SI(struct img_par *img)
{
  int i;
  const int ICBPTAB[6] = {0,16,32,15,31,47};
  Macroblock *currMB   = &img->mb_data[img->current_mb_nr];
  int         mbmode   = currMB->mb_type;

  if (mbmode==0)
  {
    currMB->mb_type = SI4MB;
    for (i=0;i<4;i++) {currMB->b8mode[i]=IBLOCK; currMB->b8pdir[i]=-1; }
    img->siblock[img->mb_x][img->mb_y]=1;
  }
  else if (mbmode==1)
  {
    currMB->mb_type = I4MB;
    for (i=0;i<4;i++) {currMB->b8mode[i]=IBLOCK; currMB->b8pdir[i]=-1; }
  }
  else
  {
    currMB->mb_type = I16MB;
    for (i=0;i<4;i++) {currMB->b8mode[i]=0; currMB->b8pdir[i]=-1; }
    currMB->cbp= ICBPTAB[(mbmode-1)>>2];
    currMB->i16mode = (mbmode-2) & 0x03;
  }
}
/*!
 ************************************************************************
 * \brief
 *    init macroblock I and P frames
 ************************************************************************
 */
void init_macroblock(struct img_par *img)
{
  int i,j;
  Macroblock *currMB = &img->mb_data[img->current_mb_nr];
  int img_block_y;
  int j2 = img->block_y/2 - 2*(img->current_mb_nr%2);

  if (img->mb_frame_field_flag)
  {
    if (img->current_mb_nr%2==0)
    {
      img_block_y   = img->block_y/2;

      img->mv_frm[img->block_x+4][img->block_y][2]=img->number;
      img->mv_top[img->block_x+4][img_block_y][2]=img->number*2;
    }
    else
    {
      img_block_y   = (img->block_y-4)/2;

      img->mv_frm[img->block_x+4][img->block_y][2]=img->number;
      img->mv_bot[img->block_x+4][img_block_y][2]=img->number*2;
    }
  }
  else
    img->mv[img->block_x+4][img->block_y][2]=img->number;

  for (i=0;i<BLOCK_SIZE;i++)
  {                           // reset vectors and pred. modes
    for(j=0;j<BLOCK_SIZE;j++)
    {
      dec_picture->mv[LIST_0][img->block_x+i][img->block_y+j][0]=0;
      dec_picture->mv[LIST_0][img->block_x+i][img->block_y+j][1]=0;
      dec_picture->mv[LIST_1][img->block_x+i][img->block_y+j][0]=0;
      dec_picture->mv[LIST_1][img->block_x+i][img->block_y+j][1]=0;

      img->ipredmode[img->block_x+i][img->block_y+j] = DC_PRED;
    }
  }

  for (j=0; j<4; j++)
    for (i=0; i<4; i++)
    {
      dec_picture->ref_idx[LIST_0][img->block_x+i][img->block_y+j] = -1;
      dec_picture->ref_idx[LIST_1][img->block_x+i][img->block_y+j] = -1;
    }
}


/*!
 ************************************************************************
 * \brief
 *    Sets mode for 8x8 block
 ************************************************************************
 */
void SetB8Mode (struct img_par* img, Macroblock* currMB, int value, int i)
{
  static const int p_v2b8 [ 5] = {4, 5, 6, 7, IBLOCK};
  static const int p_v2pd [ 5] = {0, 0, 0, 0, -1};
  static const int b_v2b8 [14] = {0, 4, 4, 4, 5, 6, 5, 6, 5, 6, 7, 7, 7, IBLOCK};
  static const int b_v2pd [14] = {2, 0, 1, 2, 0, 0, 1, 1, 2, 2, 0, 1, 2, -1};

  if (img->type==B_SLICE)
  {
    currMB->b8mode[i]   = b_v2b8[value];
    currMB->b8pdir[i]   = b_v2pd[value];

  }
  else
  {
    currMB->b8mode[i]   = p_v2b8[value];
    currMB->b8pdir[i]   = p_v2pd[value];
  }

}


void reset_coeffs()
{
  int i, j, iii, jjj;

  // reset luma coeffs
  for (i=0;i<BLOCK_SIZE;i++)
  { 
    for (j=0;j<BLOCK_SIZE;j++)
      for(iii=0;iii<BLOCK_SIZE;iii++)
        for(jjj=0;jjj<BLOCK_SIZE;jjj++)
          img->cof[i][j][iii][jjj]=0;
  }

  // reset chroma coeffs
  for (j=4;j<6;j++)
  { 
    for (i=0;i<4;i++)
      for (iii=0;iii<4;iii++)
        for (jjj=0;jjj<4;jjj++)
          img->cof[i][j][iii][jjj]=0;
  }
  
  // CAVLC
  for (i=0; i < 4; i++)
    for (j=0; j < 6; j++)
      img->nz_coeff[img->current_mb_nr][i][j]=0;  

}

/*!
 ************************************************************************
 * \brief
 *    Get the syntax elements from the NAL
 ************************************************************************
 */
int read_one_macroblock(struct img_par *img,struct inp_par *inp)
{
  int i;

  SyntaxElement currSE;
  Macroblock *currMB = &img->mb_data[img->current_mb_nr];

  Slice *currSlice = img->currentSlice;
  DataPartition *dP;
  int *partMap = assignSE2partition[currSlice->dp_mode];
  int mb_width = img->width/16;
  Macroblock *topMB = NULL;
  int  prevMbSkipped = 0;
  int  img_block_y;
  int  check_bottom, read_bottom, read_top;
  
  if (img->mb_frame_field_flag)
  {
    if (img->current_mb_nr%2)
    {
      topMB= &img->mb_data[img->current_mb_nr-1];
      if(!(img->type == B_SLICE))
        prevMbSkipped = (topMB->mb_type == 0);
      else 
        prevMbSkipped = (topMB->mb_type == 0 && topMB->cbp == 0);
    }
    else 
      prevMbSkipped = 0;
  }
  
  if (img->current_mb_nr%2 == 0)
    currMB->mb_field = 0;
  else
    currMB->mb_field = img->mb_data[img->current_mb_nr-1].mb_field;


  currMB->qp = img->qp ;

  currSE.type = SE_MBTYPE;

  //  read MB mode *****************************************************************
  if(img->type == B_SLICE) dP = &(currSlice->partArr[partMap[SE_BFRAME]]);
  else                     dP = &(currSlice->partArr[partMap[currSE.type]]);
  
  if (active_pps->entropy_coding_mode == UVLC || dP->bitstream->ei_flag)   currSE.mapping = linfo_ue;

  if(img->type == I_SLICE || img->type == SI_SLICE)
  {
    // read MB aff
    if (img->mb_frame_field_flag && img->current_mb_nr%2==0)
    {
      TRACE_STRING("mb_field_decoding_flag");
      if (active_pps->entropy_coding_mode == UVLC || dP->bitstream->ei_flag)
      {
        currSE.len = 1;
        readSyntaxElement_FLC(&currSE, dP->bitstream);
      }
      else
      {
        currSE.reading = readFieldModeInfoFromBuffer_CABAC;
        dP->readSyntaxElement(&currSE,img,inp,dP);
      }
      currMB->mb_field = currSE.value1;
    }
    
    //  read MB type
    TRACE_STRING("mb_type");
    currSE.reading = readMB_typeInfoFromBuffer_CABAC;
    dP->readSyntaxElement(&currSE,img,inp,dP);

    currMB->mb_type = currSE.value1;
    if(!dP->bitstream->ei_flag)
      currMB->ei_flag = 0;
  }

  // non I/SI-slice CABAC
  else if (active_pps->entropy_coding_mode == CABAC)
  {
    // read MB skipflag
    TRACE_STRING("mb_skip_flag");
    currSE.reading = readMB_skip_flagInfoFromBuffer_CABAC;
    dP->readSyntaxElement(&currSE,img,inp,dP);
    currMB->mb_type = currSE.value1;

    if (img->type==B_SLICE)
      currMB->cbp = currSE.value2;
    if(!dP->bitstream->ei_flag)
      currMB->ei_flag = 0;
    
    if ((img->type==B_SLICE) && currSE.value1==0 && currSE.value2==0)
      img->cod_counter=0;
      
    // read MB aff
      if (img->structure==FRAME && img->mb_frame_field_flag) 
      {
        check_bottom=read_bottom=read_top=0;
        if (img->current_mb_nr%2==0)
        {
          check_bottom =  (img->type!=B_SLICE)? 
            (currMB->mb_type == 0):
          (currMB->mb_type == 0 && currMB->cbp == 0);
          read_top = !check_bottom;
        }
        else
          read_bottom = (img->type!=B_SLICE)? 
          (topMB->mb_type == 0 && currMB->mb_type != 0) :
        ((topMB->mb_type == 0 && topMB->cbp == 0) && (currMB->mb_type != 0 || currMB->cbp != 0));
        
        if (read_bottom || read_top)
        {
          CheckAvailabilityOfNeighborsForAff(img);
          TRACE_STRING("mb_field_decoding_flag");
          currSE.reading = readFieldModeInfoFromBuffer_CABAC;
          dP->readSyntaxElement(&currSE,img,inp,dP);
          currMB->mb_field = currSE.value1;
        }
        if (check_bottom)
          currMB->mb_field = (check_next_mb_and_get_field_mode_CABAC(&currSE,img,inp,dP)) ? 0 : currSE.value1;
        
      }
      
    // read MB type
    if (currMB->mb_type != 0 )
    {
      currSE.reading = readMB_typeInfoFromBuffer_CABAC;
      TRACE_STRING("mb_type");
      dP->readSyntaxElement(&currSE,img,inp,dP);
      currMB->mb_type = currSE.value1;
      if(!dP->bitstream->ei_flag)
        currMB->ei_flag = 0;
    }
  }
  // VLC Non-Intra
  else
  {
    if(img->cod_counter == -1)
    {
      TRACE_STRING("mb_skip_run");
      dP->readSyntaxElement(&currSE,img,inp,dP);
      img->cod_counter = currSE.value1;
    }
    if (img->cod_counter==0)
    {
      // read MB aff
      if ((img->mb_frame_field_flag) && ((img->current_mb_nr%2==0) || ((img->current_mb_nr%2) && prevMbSkipped)))
      {
        TRACE_STRING("mb_field_decoding_flag");
        currSE.len = 1;
        readSyntaxElement_FLC(&currSE, dP->bitstream);
        currMB->mb_field = currSE.value1;
      }
      
      // read MB type
      TRACE_STRING("mb_type");
      dP->readSyntaxElement(&currSE,img,inp,dP);
      if(img->type == P_SLICE || img->type == SP_SLICE)
        currSE.value1++;
      currMB->mb_type = currSE.value1;
      if(!dP->bitstream->ei_flag)
        currMB->ei_flag = 0;
      img->cod_counter--;
    } 
    else
    {
      img->cod_counter--;
      currMB->mb_type = 0;
      currMB->ei_flag = 0;

      // read field flag of bottom block
      if(img->mb_frame_field_flag)
      {
        if(img->cod_counter == 0 && (img->current_mb_nr%2 == 0))
        {
          TRACE_STRING("mb_field_decoding_flag (of coded bottom mb)");
          currSE.len = 1;
          readSyntaxElement_FLC(&currSE, dP->bitstream);
          dP->bitstream->frame_bitoffset--;
          currMB->mb_field = currSE.value1;
        }
        else if(img->cod_counter > 0 && (img->current_mb_nr%2 == 0))
        {
        // check left macroblock pair first
          if (mb_is_available(img->current_mb_nr-2, img->current_mb_nr)&&((img->current_mb_nr%(img->PicWidthInMbs*2))!=0))
          {
            currMB->mb_field = img->mb_data[img->current_mb_nr-2].mb_field;
          }
          else
          {
            // check top macroblock pair
            if (mb_is_available(img->current_mb_nr-2*img->PicWidthInMbs, img->current_mb_nr))
            {
              currMB->mb_field = img->mb_data[img->current_mb_nr-2*img->PicWidthInMbs].mb_field;
            }
            else
              currMB->mb_field = 0;
          }
        }
      }
    }
  }

  dec_picture->mb_field[img->current_mb_nr] = currMB->mb_field;

  img->siblock[img->mb_x][img->mb_y]=0;

  if (img->mb_frame_field_flag)
    CheckAvailabilityOfNeighborsForAff(img);


  if ((img->type==P_SLICE ))    // inter frame
    interpret_mb_mode_P(img);
  else if (img->type==I_SLICE)                                  // intra frame
    interpret_mb_mode_I(img);
  else if ((img->type==B_SLICE))       // B frame
    interpret_mb_mode_B(img);
  else if ((img->type==SP_SLICE))     // SP frame
    interpret_mb_mode_P(img);
  else if (img->type==SI_SLICE)     // SI frame
    interpret_mb_mode_SI(img);

  if(img->mb_frame_field_flag)
  {
    if(currMB->mb_field)
    {
      img->buf_cycle = 2*(inp->buf_cycle+1);
      img->num_ref_idx_l0_active <<=1;
      img->num_ref_idx_l1_active <<=1;
    }
    else
      img->buf_cycle = inp->buf_cycle+1;
  }

  //====== READ 8x8 SUB-PARTITION MODES (modes of 8x8 blocks) and Intra VBST block modes ======
  if (IS_P8x8 (currMB))
  {
    currSE.type    = SE_MBTYPE;
    if (img->type==B_SLICE)    dP = &(currSlice->partArr[partMap[SE_BFRAME]]);
    else                       dP = &(currSlice->partArr[partMap[SE_MBTYPE]]);

    for (i=0; i<4; i++)
    {
      if (active_pps->entropy_coding_mode ==UVLC || dP->bitstream->ei_flag) currSE.mapping = linfo_ue;
      else                                                  currSE.reading = readB8_typeInfoFromBuffer_CABAC;

      TRACE_STRING("sub_mb_type");
      dP->readSyntaxElement (&currSE, img, inp, dP);
      SetB8Mode (img, currMB, currSE.value1, i);
    }
  }

  if(img->constrained_intra_pred_flag && (img->type==P_SLICE|| img->type==B_SLICE))        // inter frame
  {
    if( !IS_INTRA(currMB) )
    {
      img->intra_block[img->current_mb_nr][0] = 0;
      img->intra_block[img->current_mb_nr][1] = 0;
      img->intra_block[img->current_mb_nr][2] = 0;
      img->intra_block[img->current_mb_nr][3] = 0;
    }
  }

  //! TO for Error Concelament
  //! If we have an INTRA Macroblock and we lost the partition
  //! which contains the intra coefficients Copy MB would be better 
  //! than just a grey block.
  //! Seems to be a bit at the wrong place to do this right here, but for this case 
  //! up to now there is no other way.
  dP = &(currSlice->partArr[partMap[SE_CBP_INTRA]]);
  if(IS_INTRA (currMB) && dP->bitstream->ei_flag && img->number)
  {
    currMB->mb_type = 0;
    currMB->ei_flag = 1;
    for (i=0;i<4;i++) {currMB->b8mode[i]=currMB->b8pdir[i]=0; }
  }
  if(img->type == B_SLICE)  dP = &(currSlice->partArr[partMap[SE_BFRAME]]);
  else                    dP = &(currSlice->partArr[partMap[currSE.type]]);
  //! End TO


  //--- init macroblock data ---
  if (img->type==B_SLICE)  init_macroblock_Bframe(img);
  //  else                   
  init_macroblock       (img);

  if (IS_DIRECT (currMB) && img->cod_counter >= 0)
  {
    currMB->cbp = 0;
    reset_coeffs();

    if (active_pps->entropy_coding_mode ==CABAC)
      img->cod_counter=-1;

    return DECODE_MB;
  }

  if (IS_COPY (currMB)) //keep last macroblock
  {
    int i, j, k, pmv[2];
    int zeroMotionAbove;
    int zeroMotionLeft;
    PixelPos mb_a, mb_b;

    getLuma4x4Neighbour(img->current_mb_nr,0,0,-1, 0,&mb_a);
    getLuma4x4Neighbour(img->current_mb_nr,0,0, 0,-1,&mb_b);

    zeroMotionLeft  = !mb_a.available ? 1 : dec_picture->ref_idx[LIST_0][mb_a.pos_x][mb_a.pos_y]==0 && dec_picture->mv[LIST_0][mb_a.pos_x][mb_a.pos_y][0]==0 && dec_picture->mv[LIST_0][mb_a.pos_x][mb_a.pos_y][1]==0 ? 1 : 0;
    zeroMotionAbove = !mb_b.available ? 1 : dec_picture->ref_idx[LIST_0][mb_b.pos_x][mb_b.pos_y]==0 && dec_picture->mv[LIST_0][mb_b.pos_x][mb_b.pos_y][0]==0 && dec_picture->mv[LIST_0][mb_b.pos_x][mb_b.pos_y][1]==0 ? 1 : 0;

    currMB->cbp = 0;
    reset_coeffs();

    img_block_y   = img->block_y;
  
    if (zeroMotionAbove || zeroMotionLeft)
    {
      for(i=0;i<BLOCK_SIZE;i++)
        for(j=0;j<BLOCK_SIZE;j++)
          for (k=0;k<2;k++)
            dec_picture->mv[LIST_0][img->block_x+i][img->block_y+j][k] = 0;
    }
    else
    {
      SetMotionVectorPredictor (img, pmv, pmv+1, 0, LIST_0, dec_picture->ref_idx, dec_picture->mv, 0, 0, 16, 16);
      
      for(i=0;i<BLOCK_SIZE;i++)
        for(j=0;j<BLOCK_SIZE;j++)
          for (k=0;k<2;k++)
          {
            dec_picture->mv[LIST_0][img->block_x+i][img_block_y+j][k] = pmv[k];
          }
    }

    for(i=0;i<BLOCK_SIZE;i++)
      for(j=0;j<BLOCK_SIZE;j++)
      {
        dec_picture->ref_idx[LIST_0][img->block_x+i][img_block_y+j] = 0;
      }

    return DECODE_MB;
  }

  // intra prediction modes for a macroblock 4x4 **********************************************
  read_ipred_modes(img,inp);

  // read inter frame vector data *********************************************************
  if (IS_INTERMV (currMB))
  {
    readMotionInfoFromNAL (img, inp);
  }

  // read CBP and Coeffs  ***************************************************************
  readCBPandCoeffsFromNAL (img,inp);

  return DECODE_MB;
}

void read_ipred_modes(struct img_par *img,struct inp_par *inp)
{
  int b8,i,j,bi,bj,bx,by,dec;
  SyntaxElement currSE;
  Slice *currSlice;
  DataPartition *dP;
  int *partMap;
  Macroblock *currMB;
  int ts, ls;
  int mostProbableIntraPredMode;
  int upIntraPredMode;
  int leftIntraPredMode;
  int IntraChromaPredModeFlag;

  PixelPos left_block;

  PixelPos top_block;
  
  currMB = &img->mb_data[img->current_mb_nr];

  IntraChromaPredModeFlag = IS_INTRA(currMB);

  currSlice = img->currentSlice;
  partMap = assignSE2partition[currSlice->dp_mode];

  currSE.type = SE_INTRAPREDMODE;

  TRACE_STRING("intra4x4_pred_mode");
  if(img->type == B_SLICE)     dP = &(currSlice->partArr[partMap[SE_BFRAME]]);
  else                         dP = &(currSlice->partArr[partMap[currSE.type]]);

  if (!(active_pps->entropy_coding_mode == UVLC || dP->bitstream->ei_flag)) 
    currSE.reading = readIntraPredModeFromBuffer_CABAC;

  for(b8=0;b8<4;b8++)  //loop 8x8 blocks
  {
    if( currMB->b8mode[b8]==IBLOCK )
    {
      IntraChromaPredModeFlag = 1;

      for(j=0;j<2;j++)  //loop subblocks
        for(i=0;i<2;i++)
        {
          //get from stream
          if (active_pps->entropy_coding_mode == UVLC || dP->bitstream->ei_flag)
            readSyntaxElement_Intra4x4PredictionMode(&currSE,img,inp,dP);
          else 
          {
            currSE.context=(b8<<2)+(j<<1)+i;
            dP->readSyntaxElement(&currSE,img,inp,dP);
          }

          bx = ((b8&1)<<1) + i;
          by = (b8&2)      + j;

          getLuma4x4Neighbour(img->current_mb_nr, bx, by, -1,  0, &left_block);
          getLuma4x4Neighbour(img->current_mb_nr, bx, by,  0, -1, &top_block);
          
          //get from array and decode
          bi = img->block_x + bx;
          bj = img->block_y + by;

          // !! KS: not sure if the follwing is still correct...
          ts=ls=0;   // Check to see if the neighboring block is SI
          if (IS_OLDINTRA(currMB) && img->type == SI_SLICE)           // need support for MBINTLC1
          {
            if (left_block.available)
              if (img->siblock [left_block.pos_x][left_block.pos_y])
                ls=1;

            if (top_block.available)
              if (img->siblock [top_block.pos_x][top_block.pos_y])
                ts=1;
          }

          upIntraPredMode            = (top_block.available  &&(ts == 0)) ? img->ipredmode[top_block.pos_x ][top_block.pos_y ] : -1;
          leftIntraPredMode          = (left_block.available &&(ls == 0)) ? img->ipredmode[left_block.pos_x][left_block.pos_y] : -1;

          mostProbableIntraPredMode  = (upIntraPredMode < 0 || leftIntraPredMode < 0) ? DC_PRED : upIntraPredMode < leftIntraPredMode ? upIntraPredMode : leftIntraPredMode;

          dec = (currSE.value1 == -1) ? mostProbableIntraPredMode : currSE.value1 + (currSE.value1 >= mostProbableIntraPredMode);

          //set
          img->ipredmode[bi][bj]=dec;
        }
    }
  }

  if (IntraChromaPredModeFlag)
  {
    currSE.type = SE_INTRAPREDMODE;
    TRACE_STRING("intra_chroma_pred_mode");
    if(img->type == B_SLICE)     dP = &(currSlice->partArr[partMap[SE_BFRAME]]);
    else                         dP = &(currSlice->partArr[partMap[currSE.type]]);

    if (active_pps->entropy_coding_mode == UVLC || dP->bitstream->ei_flag) currSE.mapping = linfo_ue;
    else                                                    currSE.reading = readCIPredMode_FromBuffer_CABAC;

    dP->readSyntaxElement(&currSE,img,inp,dP);
    currMB->c_ipred_mode = currSE.value1;

    if (currMB->c_ipred_mode < DC_PRED_8 || currMB->c_ipred_mode > PLANE_8)
    {
      error("illegal chroma intra pred mode!\n", 600);
    }
  }
}



/*!
 ************************************************************************
 * \brief
 *    Set motion vector predictor
 ************************************************************************
 */
static void SetMotionVectorPredictor (struct img_par  *img,
                                      int             *pmv_x,
                                      int             *pmv_y,
                                      int             ref_frame,
                                      int             list,
                                      int             ***refPic,
                                      int             ****tmp_mv,
                                      int             block_x,
                                      int             block_y,
                                      int             blockshape_x,
                                      int             blockshape_y)
{
  int mb_x                 = 4*block_x;
  int mb_y                 = 4*block_y;
  int pic_block_x          = img->block_x + block_x;
  int pic_block_y          = img->block_y + block_y;
  int mb_width             = img->width/16;
  int mb_nr                = img->current_mb_nr;

  int mv_a, mv_b, mv_c, pred_vec=0;
  int mvPredType, rFrameL, rFrameU, rFrameUR;
  int hv;


  PixelPos block_a, block_b, block_c, block_d;

  getLuma4x4Neighbour(mb_nr, block_x, block_y,           -1,  0, &block_a);
  getLuma4x4Neighbour(mb_nr, block_x, block_y,            0, -1, &block_b);
  getLuma4x4Neighbour(mb_nr, block_x, block_y, blockshape_x, -1, &block_c);
  getLuma4x4Neighbour(mb_nr, block_x, block_y,           -1, -1, &block_d);

  if (mb_y > 0)
  {
    if (mb_x < 8)  // first column of 8x8 blocks
    {
      if (mb_y==8)
      {
        if (blockshape_x == 16)      block_c.available  = 0;
        else                         block_c.available &= 1;
      }
      else
      {
        if (mb_x+blockshape_x != 8)  block_c.available &= 1;
        else                         block_c.available  = 0;
      }
    }
    else
    {
      if (mb_x+blockshape_x != 16)   block_c.available &= 1;
      else                           block_c.available  = 0;
    }
  }

  if (!block_c.available)
  {
    block_c=block_d;
  }

  mvPredType = MVPRED_MEDIAN;

  if (!img->mb_frame_field_flag)
  {
    rFrameL    = block_a.available    ? refPic[list][block_a.pos_x][block_a.pos_y] : -1;
    rFrameU    = block_b.available    ? refPic[list][block_b.pos_x][block_b.pos_y] : -1;
    rFrameUR   = block_c.available    ? refPic[list][block_c.pos_x][block_c.pos_y] : -1;
  }
  else
  {
    if (img->mb_data[img->current_mb_nr].mb_field)
    {
      rFrameL    = block_a.available    ? 
        img->mb_data[block_a.mb_addr].mb_field ? 
        refPic[list][block_a.pos_x][block_a.pos_y]:
        refPic[list][block_a.pos_x][block_a.pos_y] * 2: 
        -1;
      rFrameU    = block_b.available    ? 
        img->mb_data[block_b.mb_addr].mb_field ? 
        refPic[list][block_b.pos_x][block_b.pos_y]:
        refPic[list][block_b.pos_x][block_b.pos_y] * 2: 
        -1;
      rFrameUR    = block_c.available    ? 
        img->mb_data[block_c.mb_addr].mb_field ? 
        refPic[list][block_c.pos_x][block_c.pos_y]:
        refPic[list][block_c.pos_x][block_c.pos_y] * 2: 
        -1;
    }
    else
    {
      rFrameL    = block_a.available    ? 
        img->mb_data[block_a.mb_addr].mb_field ? 
        refPic[list][block_a.pos_x][block_a.pos_y] >>1:
        refPic[list][block_a.pos_x][block_a.pos_y] : 
        -1;
      rFrameU    = block_b.available    ? 
        img->mb_data[block_b.mb_addr].mb_field ? 
        refPic[list][block_b.pos_x][block_b.pos_y] >>1:
        refPic[list][block_b.pos_x][block_b.pos_y] : 
        -1;
      rFrameUR    = block_c.available    ? 
        img->mb_data[block_c.mb_addr].mb_field ? 
        refPic[list][block_c.pos_x][block_c.pos_y] >>1:
        refPic[list][block_c.pos_x][block_c.pos_y] : 
        -1;
    }
  }


  /* Prediction if only one of the neighbors uses the reference frame
   * we are checking
   */
  if(rFrameL == ref_frame && rFrameU != ref_frame && rFrameUR != ref_frame)       mvPredType = MVPRED_L;
  else if(rFrameL != ref_frame && rFrameU == ref_frame && rFrameUR != ref_frame)  mvPredType = MVPRED_U;
  else if(rFrameL != ref_frame && rFrameU != ref_frame && rFrameUR == ref_frame)  mvPredType = MVPRED_UR;
  // Directional predictions 
  else if(blockshape_x == 8 && blockshape_y == 16)
  {
    if(mb_x == 0)
    {
      if(rFrameL == ref_frame)
        mvPredType = MVPRED_L;
    }
    else
    {
      if( rFrameUR == ref_frame)
        mvPredType = MVPRED_UR;
    }
  }
  else if(blockshape_x == 16 && blockshape_y == 8)
  {
    if(mb_y == 0)
    {
      if(rFrameU == ref_frame)
        mvPredType = MVPRED_U;
    }
    else
    {
      if(rFrameL == ref_frame)
        mvPredType = MVPRED_L;
    }
  }

  for (hv=0; hv < 2; hv++)
  {
    if (!img->mb_frame_field_flag || hv==0)
    {
      mv_a = block_a.available  ? tmp_mv[list][block_a.pos_x][block_a.pos_y][hv] : 0;
      mv_b = block_b.available  ? tmp_mv[list][block_b.pos_x][block_b.pos_y][hv] : 0;
      mv_c = block_c.available  ? tmp_mv[list][block_c.pos_x][block_c.pos_y][hv] : 0;
    }
    else
    {
      if (img->mb_data[img->current_mb_nr].mb_field)
      {
        mv_a = block_a.available  ? img->mb_data[block_a.mb_addr].mb_field?
          tmp_mv[list][block_a.pos_x][block_a.pos_y][hv]:
          tmp_mv[list][block_a.pos_x][block_a.pos_y][hv] / 2: 
          0;
        mv_b = block_b.available  ? img->mb_data[block_b.mb_addr].mb_field?
          tmp_mv[list][block_b.pos_x][block_b.pos_y][hv]:
          tmp_mv[list][block_b.pos_x][block_b.pos_y][hv] / 2: 
          0;
        mv_c = block_c.available  ? img->mb_data[block_c.mb_addr].mb_field?
          tmp_mv[list][block_c.pos_x][block_c.pos_y][hv]:
          tmp_mv[list][block_c.pos_x][block_c.pos_y][hv] / 2: 
          0;
      }
      else
      {
        mv_a = block_a.available  ? img->mb_data[block_a.mb_addr].mb_field?
          tmp_mv[list][block_a.pos_x][block_a.pos_y][hv] * 2:
          tmp_mv[list][block_a.pos_x][block_a.pos_y][hv]: 
          0;
        mv_b = block_b.available  ? img->mb_data[block_b.mb_addr].mb_field?
          tmp_mv[list][block_b.pos_x][block_b.pos_y][hv] * 2:
          tmp_mv[list][block_b.pos_x][block_b.pos_y][hv]: 
          0;
        mv_c = block_c.available  ? img->mb_data[block_c.mb_addr].mb_field?
          tmp_mv[list][block_c.pos_x][block_c.pos_y][hv] * 2:
          tmp_mv[list][block_c.pos_x][block_c.pos_y][hv]: 
          0;
      }
    }


    switch (mvPredType)
    {
    case MVPRED_MEDIAN:
      if(!(block_b.available || block_c.available))
        pred_vec = mv_a;
      else
        pred_vec = mv_a+mv_b+mv_c-min(mv_a,min(mv_b,mv_c))-max(mv_a,max(mv_b,mv_c));
      break;
    case MVPRED_L:
      pred_vec = mv_a;
      break;
    case MVPRED_U:
      pred_vec = mv_b;
      break;
    case MVPRED_UR:
      pred_vec = mv_c;
      break;
    default:
      break;
    }

    if (hv==0)  *pmv_x = pred_vec;
    else        *pmv_y = pred_vec;

  }
}


/*!
 ************************************************************************
 * \brief
 *    Set context for reference frames
 ************************************************************************
 */
int
BType2CtxRef (int btype)
{
  if (btype<4)  return 0;
  else          return 1;
}

/*!
 ************************************************************************
 * \brief
 *    Read motion info
 ************************************************************************
 */
void readMotionInfoFromNAL (struct img_par *img, struct inp_par *inp)
{
  int i,j,k;
  int step_h,step_v;
  int curr_mvd;
  Macroblock *currMB  = &img->mb_data[img->current_mb_nr];
  SyntaxElement currSE;
  Slice *currSlice    = img->currentSlice;
  DataPartition *dP;
  int *partMap        = assignSE2partition[currSlice->dp_mode];
  int bframe          = (img->type==B_SLICE);
  int partmode        = (IS_P8x8(currMB)?4:currMB->mb_type);
  int step_h0         = BLOCK_STEP [partmode][0];
  int step_v0         = BLOCK_STEP [partmode][1];

  int mv_mode, i0, j0, refframe;
  int pmv[2];
  int j4, i4, ii,jj;
  int vec;

  int iTRb,iTRp;
  int mv_scale;
  int frame_no_next_P, frame_no_B, delta_P;
  int ref;
  int img_block_y;
//  int use_scaled_mv;
//  int refframe_l0,current_tr;

  int **fwRefFrArr = img->fw_refFrArr;
  int **bwRefFrArr = img->bw_refFrArr;
  int  ***fw_mv = img->fw_mv;
  int  ***bw_mv = img->bw_mv;
  int  **moving_block_dir = moving_block; 
  int  ***fw_mv_array, ***bw_mv_array;
  int j6;  
  int flag_mode;

  if (bframe && IS_P8x8 (currMB))
  {
    if (img->direct_type && img->mb_frame_field_flag )
    {
      if (!currMB->mb_field)
      {
        fwRefFrArr= img->fw_refFrArr_frm;
        bwRefFrArr= img->bw_refFrArr_frm;
        fw_mv=img->fw_mv_frm;
        bw_mv=img->bw_mv_frm;
        fw_mv_array = img->dfMV;
        bw_mv_array = img->dbMV;
      }
      else if (img->current_mb_nr%2 )
      {
        fwRefFrArr= img->fw_refFrArr_bot;
        bwRefFrArr= img->bw_refFrArr_bot;
        fw_mv=img->fw_mv_bot;
        bw_mv=img->bw_mv_bot;
        moving_block_dir = moving_block_bot; 
        fw_mv_array = img->dfMV_bot;
        bw_mv_array = img->dbMV_bot;
      }
      else
      {
        fwRefFrArr=img->fw_refFrArr_top ;
        bwRefFrArr=img->bw_refFrArr_top ;
        fw_mv=img->fw_mv_top;
        bw_mv=img->bw_mv_top;
        moving_block_dir = moving_block_top; 
        fw_mv_array = img->dfMV_top;
        bw_mv_array = img->dbMV_top;
      }
    }
    if (img->direct_type)
    {
      int pic_blockx          = img->block_x;
      int pic_blocky          = (img->mb_frame_field_flag && currMB->mb_field)? ((img->current_mb_nr%2)?img->block_y/2-BLOCK_SIZE / 2:img->block_y/2):img->block_y;
      int mb_nr                = img->current_mb_nr;
      int mb_width             = img->width/16;
      int mb_available_up =      (img->mb_y == 0 || pic_blocky == 0  ) ? 0 : (img->mb_frame_field_flag? 1 :(currMB->slice_nr == img->mb_data[mb_nr-mb_width].slice_nr));
      int mb_available_left =    (img->mb_x == 0                      ) ? 0 : (currMB->slice_nr == img->mb_data[mb_nr-1].slice_nr);
      int mb_available_upleft  = (img->mb_x == 0 || img->mb_y == 0 || pic_blocky == 0) ? 0 : (img->mb_frame_field_flag)? 1 :(currMB->slice_nr == img->mb_data[mb_nr-mb_width-1].slice_nr);
      int mb_available_upright = (img->mb_frame_field_flag && img->current_mb_nr%2)?0:(img->mb_x >= mb_width-1 ||
        img->mb_y == 0 || pic_blocky == 0 ) ? 0 : (img->mb_frame_field_flag)? 1 :(currMB->slice_nr == img->mb_data[mb_nr-mb_width+1].slice_nr);
      
      
      int fw_rFrameL              = mb_available_left    ? fwRefFrArr[pic_blocky]  [pic_blockx-1]   : -1;
      int fw_rFrameU              = mb_available_up      ? fwRefFrArr[pic_blocky-1][pic_blockx]     : -1;
      int fw_rFrameUL             = mb_available_upleft  ? fwRefFrArr[pic_blocky-1][pic_blockx-1]   : -1;
      int fw_rFrameUR             = mb_available_upright ? fwRefFrArr[pic_blocky-1][pic_blockx+4]   : fw_rFrameUL;  
      
      int bw_rFrameL              = mb_available_left    ? bwRefFrArr[pic_blocky]  [pic_blockx-1]   : -1;
      int bw_rFrameU              = mb_available_up      ? bwRefFrArr[pic_blocky-1][pic_blockx]     : -1;
      int bw_rFrameUL             = mb_available_upleft  ? bwRefFrArr[pic_blocky-1][pic_blockx-1]   : -1;
      int bw_rFrameUR             = mb_available_upright ? bwRefFrArr[pic_blocky-1][pic_blockx+4]   : bw_rFrameUL;
      
      
      int fw_rFrame,bw_rFrame;
      int pmvfw[2]={0,0},pmvbw[2]={0,0};
      int j5=0;
      
      if (!fw_rFrameL || !fw_rFrameU || !fw_rFrameUR)
        fw_rFrame=0;
      else
        fw_rFrame=min(fw_rFrameL&15,min(fw_rFrameU&15,fw_rFrameUR&15));
      
      if(img->num_ref_idx_l1_active>1 && (bw_rFrameL==1 || bw_rFrameU==1 || bw_rFrameUR==1))
        bw_rFrame=1;
      else if (!bw_rFrameL || !bw_rFrameU || !bw_rFrameUR)
        bw_rFrame=0;
      else
        bw_rFrame=min(bw_rFrameL&15,min(bw_rFrameU&15,bw_rFrameUR&15));
      
      if (fw_rFrame !=15)
/*        SetMotionVectorPredictor (img, pmvfw, pmvfw+1, fw_rFrame, fwRefFrArr, fw_mv, 0, 0, 16, 16) */;  
      if (bw_rFrame !=15)
/*        SetMotionVectorPredictor (img, pmvbw, pmvbw+1, bw_rFrame, bwRefFrArr, bw_mv, 0, 0, 16, 16) */;  
      
      for (i=0;i<4;i++)
      {
        if (currMB->b8mode[i] == 0)
          for(j=2*(i/2);j<2*(i/2)+2;j++)
            for(k=2*(i%2);k<2*(i%2)+2;k++)
            {
              j4 = img->block_y+j;
              j6 = pic_blocky+j;
              i4 = img->block_x+k;
              if (img->mb_frame_field_flag) 
              {
                j5 = img->block_y / 2 + j;
                if (img->current_mb_nr%2)
                  j5 -= BLOCK_SIZE / 2;
              }
              if (!(img->mb_frame_field_flag && currMB->mb_field))
              {
                if (fw_rFrame !=15)
                {
                  if  (!fw_rFrame  && !moving_block_dir[j6][i4])       
                  {                    
                    img->fw_mv[i4+BLOCK_SIZE][j4][0]=img->dfMV[i4+BLOCK_SIZE][j4][0] = 0;
                    img->fw_mv[i4+BLOCK_SIZE][j4][1]=img->dfMV[i4+BLOCK_SIZE][j4][1] = 0;
                    if (img->mb_frame_field_flag)
                    {
                      if (img->current_mb_nr%2 == 0)
                      {
                        img->dfMV_top[i4+BLOCK_SIZE][j5][0]=img->fw_mv_top[i4+BLOCK_SIZE][j5][0]=0;
                        img->dfMV_top[i4+BLOCK_SIZE][j5][1]=img->fw_mv_top[i4+BLOCK_SIZE][j5][1]=0;
                        img->fw_refFrArr_top[j5][i4]=0;                        
                      }
                      else
                      {
                        img->dfMV_bot[i4+BLOCK_SIZE][j5][0]=img->fw_mv_bot[i4+BLOCK_SIZE][j5][0]=0;
                        img->dfMV_bot[i4+BLOCK_SIZE][j5][1]=img->fw_mv_bot[i4+BLOCK_SIZE][j5][1]=0;
                        img->fw_refFrArr_bot[j5][i4]=0;
                      }
                    }                    
                    if (img->structure == TOP_FIELD)  //! Note that this seems to be unecessary for different img->structure.
                      fwRefFrArr[j4][i4] = 0;         //! copied to be consistent with temporal direct, just in case there is 
                    else                              //! change in the software.
                      fwRefFrArr[j4][i4] = 0;                                      
                  }
                  else
                  {
                    img->fw_mv[i4+BLOCK_SIZE][j4][0]=img->dfMV[i4+BLOCK_SIZE][j4][0] = pmvfw[0];
                    img->fw_mv[i4+BLOCK_SIZE][j4][1]=img->dfMV[i4+BLOCK_SIZE][j4][1] = pmvfw[1];
                    if (img->structure == TOP_FIELD)
                      fwRefFrArr[j4][i4] = fw_rFrame ;
                    else                
                      fwRefFrArr[j4][i4] = fw_rFrame;                
                    if (img->mb_frame_field_flag)
                    {
                      if (img->current_mb_nr%2 == 0)
                      {
                        img->dfMV_top[i4+BLOCK_SIZE][j5][0]=img->fw_mv_top[i4+BLOCK_SIZE][j5][0]=pmvfw[0];
                        img->dfMV_top[i4+BLOCK_SIZE][j5][1]=img->fw_mv_top[i4+BLOCK_SIZE][j5][1]=pmvfw[1]/2;
                        img->fw_refFrArr_top[j5][i4]=2*fw_rFrame;
                      }
                      else
                      {
                        img->dfMV_bot[i4+BLOCK_SIZE][j5][0]=img->fw_mv_bot[i4+BLOCK_SIZE][j5][0]=pmvfw[0];
                        img->dfMV_bot[i4+BLOCK_SIZE][j5][1]=img->fw_mv_bot[i4+BLOCK_SIZE][j5][1]=pmvfw[1]/2;
                        img->fw_refFrArr_bot[j5][i4]=2*fw_rFrame;
                      }
                    }
                  }
                }
                else 
                {
                  img->fw_refFrArr[j4][i4]=-1;                                   
                  img->fw_mv[i4+BLOCK_SIZE][j4][0]=img->dfMV[i4+BLOCK_SIZE][j4][0] = 0;
                  img->fw_mv[i4+BLOCK_SIZE][j4][1]=img->dfMV[i4+BLOCK_SIZE][j4][1] = 0;
                  if (img->mb_frame_field_flag)
                  {
                    if (img->current_mb_nr%2 == 0)
                    {
                      img->dfMV_top[i4+BLOCK_SIZE][j5][0]=img->fw_mv_top[i4+BLOCK_SIZE][j5][0]=0;
                      img->dfMV_top[i4+BLOCK_SIZE][j5][1]=img->fw_mv_top[i4+BLOCK_SIZE][j5][1]=0;
                      img->fw_refFrArr_top[j5][i4]=-1;
                    }
                    else
                    {
                      img->dfMV_bot[i4+BLOCK_SIZE][j5][0]=img->fw_mv_bot[i4+BLOCK_SIZE][j5][0]=0;
                      img->dfMV_bot[i4+BLOCK_SIZE][j5][1]=img->fw_mv_bot[i4+BLOCK_SIZE][j5][1]=0;
                      img->fw_refFrArr_bot[j5][i4]=-1;
                    }
                  }
                }
                if (bw_rFrame !=15)
                {
                  if  (bw_rFrame==((img->num_ref_idx_l1_active>1)?1:0) && !moving_block_dir[j6][i4])                         
                  {                  
                    img->bw_mv[i4+BLOCK_SIZE][j4][0]=img->dbMV[i4+BLOCK_SIZE][j4][0] = 0;
                    img->bw_mv[i4+BLOCK_SIZE][j4][1]=img->dbMV[i4+BLOCK_SIZE][j4][1] = 0;
                    bwRefFrArr[j4][i4]=bw_rFrame;
                    if (img->mb_frame_field_flag)
                    {
                      if (img->current_mb_nr%2 == 0)
                      {
                        img->dbMV_top[i4+BLOCK_SIZE][j5][0]=img->bw_mv_top[i4+BLOCK_SIZE][j5][0]=0;
                        img->dbMV_top[i4+BLOCK_SIZE][j5][1]=img->bw_mv_top[i4+BLOCK_SIZE][j5][1]=0;
                        img->bw_refFrArr_top[j5][i4]=2*bw_rFrame;
                      }
                      else
                      {
                        img->dbMV_bot[i4+BLOCK_SIZE][j5][0]=img->bw_mv_bot[i4+BLOCK_SIZE][j5][0]=0;
                        img->dbMV_bot[i4+BLOCK_SIZE][j5][1]=img->bw_mv_bot[i4+BLOCK_SIZE][j5][1]=0;
                        img->bw_refFrArr_bot[j5][i4]=2*bw_rFrame;
                      }
                    }
                  }
                  else
                  {                    
                    img->bw_mv[i4+BLOCK_SIZE][j4][0]=img->dbMV[i4+BLOCK_SIZE][j4][0] = pmvbw[0];
                    img->bw_mv[i4+BLOCK_SIZE][j4][1]=img->dbMV[i4+BLOCK_SIZE][j4][1] = pmvbw[1];
                    bwRefFrArr[j4][i4]=bw_rFrame;
                    if (img->mb_frame_field_flag)
                    {
                      if (img->current_mb_nr%2 == 0)
                      {
                        img->dbMV_top[i4+BLOCK_SIZE][j5][0]=img->bw_mv_top[i4+BLOCK_SIZE][j5][0]=pmvbw[0];
                        img->dbMV_top[i4+BLOCK_SIZE][j5][1]=img->bw_mv_top[i4+BLOCK_SIZE][j5][1]=pmvbw[1]/2;
                        img->bw_refFrArr_top[j5][i4]=2*bw_rFrame;
                      }
                      else
                      {
                        img->dbMV_bot[i4+BLOCK_SIZE][j5][0]=img->bw_mv_bot[i4+BLOCK_SIZE][j5][0]=pmvbw[0];
                        img->dbMV_bot[i4+BLOCK_SIZE][j5][1]=img->bw_mv_bot[i4+BLOCK_SIZE][j5][1]=pmvbw[1]/2;
                        img->bw_refFrArr_bot[j5][i4]=2*bw_rFrame;
                      }
                    }
                  }
                }
                else
                {                                                            
                  img->bw_mv[i4+BLOCK_SIZE][j4][0]=img->dbMV[i4+BLOCK_SIZE][j4][0] = 0;
                  img->bw_mv[i4+BLOCK_SIZE][j4][1]=img->dbMV[i4+BLOCK_SIZE][j4][1] = 0;
                  bwRefFrArr[j4][i4]=-1;                  
                  
                  if (img->mb_frame_field_flag)
                  {
                    if (img->current_mb_nr%2 == 0)
                    {
                      img->dbMV_top[i4+BLOCK_SIZE][j5][0]=img->bw_mv_top[i4+BLOCK_SIZE][j5][0]=0;
                      img->dbMV_top[i4+BLOCK_SIZE][j5][1]=img->bw_mv_top[i4+BLOCK_SIZE][j5][1]=0;
                      img->bw_refFrArr_top[j5][i4]=-1;
                    }
                    else
                    {
                      img->dbMV_bot[i4+BLOCK_SIZE][j5][0]=img->bw_mv_bot[i4+BLOCK_SIZE][j5][0]=0;
                      img->dbMV_bot[i4+BLOCK_SIZE][j5][1]=img->bw_mv_bot[i4+BLOCK_SIZE][j5][1]=0;
                      img->bw_refFrArr_bot[j5][i4]=-1;
                    }
                  }
                }
                if (fw_rFrame ==15 && bw_rFrame ==15)
                {
                  if (img->structure == TOP_FIELD)
                    fwRefFrArr[j4][i4] =  0;
                  else                
                    fwRefFrArr[j4][i4]=0;                
                  bwRefFrArr[j4][i4] =(img->num_ref_idx_l1_active>1)?1:0;
                  
                  if (img->mb_frame_field_flag)
                  {
                    if (img->current_mb_nr%2 == 0)
                    {
                      img->fw_refFrArr_top[j5][i4]=0;
                      img->bw_refFrArr_top[j5][i4]=2*bwRefFrArr[j4][i4];
                    }
                    else
                    {
                      img->fw_refFrArr_bot[j5][i4]=0;                        
                      img->bw_refFrArr_bot[j5][i4]=2*bwRefFrArr[j4][i4];
                    }
                  }
                }
                
                if (img->mb_frame_field_flag)
                {
                  
                  img->fw_mv_frm[i4+BLOCK_SIZE][j4][0]=img->fw_mv[i4+BLOCK_SIZE][j4][0];
                  img->fw_mv_frm[i4+BLOCK_SIZE][j4][1]=img->fw_mv[i4+BLOCK_SIZE][j4][1];
                  img->bw_mv_frm[i4+BLOCK_SIZE][j4][0]=img->bw_mv[i4+BLOCK_SIZE][j4][0];
                  img->bw_mv_frm[i4+BLOCK_SIZE][j4][1]=img->bw_mv[i4+BLOCK_SIZE][j4][1];
                  if (currMB->mb_field)
                  {
                    img->fw_refFrArr_frm[img->block_y+j][img->block_x+i] = (fwRefFrArr[j4][i4]==-1) ? -1 : (fwRefFrArr[j4][i4] + 1)/2;
                    img->bw_refFrArr_frm[img->block_y+j][img->block_x+i] = (bwRefFrArr[j4][i4]==-1) ? -1 : (bwRefFrArr[j4][i4] + 1)/2;
                    // Revisit Alexis   img->bw_refFrArr_frm[img->block_y + j][img->block_x + k] = -1;
                  } 
                }
              }
              else //! (img->mb_frame_field_flag && img->mb_field)
              {
                if (fw_rFrame !=15)
                {
                  if  (!fw_rFrame  && !moving_block_dir[j5][i4])       
                  {                    
                    fwRefFrArr[j5][i4] = 0;
                    fw_mv_array[i4+BLOCK_SIZE][j5][0]=fw_mv[i4+BLOCK_SIZE][j5][0]=0;
                    fw_mv_array[i4+BLOCK_SIZE][j5][1]=fw_mv[i4+BLOCK_SIZE][j5][1]=0;
                    
                  }
                  else
                  {
                    fwRefFrArr[j5][i4] = fw_rFrame ;                    
                    fw_mv_array[i4+BLOCK_SIZE][j5][0]=fw_mv[i4+BLOCK_SIZE][j5][0]=pmvfw[0];
                    fw_mv_array[i4+BLOCK_SIZE][j5][1]=fw_mv[i4+BLOCK_SIZE][j5][1]=pmvfw[1];                    
                  }
                }
                else
                {
                  fwRefFrArr[j5][i4]=-1;
                  fw_mv_array[i4+BLOCK_SIZE][j5][0]=fw_mv[i4+BLOCK_SIZE][j5][0]=0;
                  fw_mv_array[i4+BLOCK_SIZE][j5][1]=fw_mv[i4+BLOCK_SIZE][j5][1]=0;
                }
                if (bw_rFrame !=15)
                {
                  
                  if  (bw_rFrame==((img->num_ref_idx_l1_active>1)?1:0) && !moving_block_dir[j5][i4])                         
                  {
                    bwRefFrArr[j5][i4]=bw_rFrame;
                    bw_mv_array[i4+BLOCK_SIZE][j5][0]=bw_mv[i4+BLOCK_SIZE][j5][0]=0;
                    bw_mv_array[i4+BLOCK_SIZE][j5][1]=bw_mv[i4+BLOCK_SIZE][j5][1]=0;
                  }
                  else
                  {
                    bwRefFrArr[j5][i4]=bw_rFrame;
                    bw_mv_array[i4+BLOCK_SIZE][j5][0]=bw_mv[i4+BLOCK_SIZE][j5][0]=pmvbw[0];
                    bw_mv_array[i4+BLOCK_SIZE][j5][1]=bw_mv[i4+BLOCK_SIZE][j5][1]=pmvbw[1];
                  }
                }
                else
                {
                  bwRefFrArr[j5][i4]=-1;
                  bw_mv_array[i4+BLOCK_SIZE][j5][0]=bw_mv[i4+BLOCK_SIZE][j5][0]=0;
                  bw_mv_array[i4+BLOCK_SIZE][j5][1]=bw_mv[i4+BLOCK_SIZE][j5][1]=0;
                }
                if (fw_rFrame ==15 && bw_rFrame ==15)
                {
                  fwRefFrArr[j5][i4]=  0;
                  bwRefFrArr[j5][i4] =(img->num_ref_idx_l1_active>1)?1:0;
                }
                if (fwRefFrArr[j5][i4]!=-1)
                  img->fw_refFrArr_frm[j4][i4]=fwRefFrArr[j5][i4]/2;
                else
                  img->fw_refFrArr_frm[j4][i4]=-1;
                
                if (bwRefFrArr[j5][i4]!=-1)
                  img->bw_refFrArr_frm[j4][i4]=bwRefFrArr[j5][i4]/2;
                else
                  img->bw_refFrArr_frm[j4][i4]=-1;

                img->fw_mv_frm[i4+BLOCK_SIZE][j4][0]=fw_mv[i4+BLOCK_SIZE][j5][0];
                img->fw_mv_frm[i4+BLOCK_SIZE][j4][1]=fw_mv[i4+BLOCK_SIZE][j5][1]*2;
                img->bw_mv_frm[i4+BLOCK_SIZE][j4][0]=bw_mv[i4+BLOCK_SIZE][j5][0];
                img->bw_mv_frm[i4+BLOCK_SIZE][j4][1]=bw_mv[i4+BLOCK_SIZE][j5][1]*2;
              }
            }
        }
    }
    else
    {
      for (i=0;i<4;i++)
      {
        if (currMB->b8mode[i] == 0)
        {
          for(j=2*(i/2);j<2*(i/2)+2;j++)
          {
            for(k=2*(i%2);k<2*(i%2)+2;k++)
            {
              if (img->mb_frame_field_flag && img->structure==FRAME)
              {
                if (currMB->mb_field==0)
                {
                  if (refFrArr_frm[img->block_y+j][img->block_x+k] == -1)
                  {
                    img->fw_refFrArr_frm[img->block_y + j][img->block_x + k] = -1;
                    img->bw_refFrArr_frm[img->block_y + j][img->block_x + k] = -1;
                  }
                  else
                  {
                    img->fw_refFrArr_frm[img->block_y + j][img->block_x + k] = refFrArr_frm[img->block_y+j][img->block_x+k];
                    img->bw_refFrArr_frm[img->block_y + j][img->block_x + k] = 0;
                  }
                }
                else if (img->current_mb_nr%2==0)
                {
                  img_block_y   = img->block_y/2;
                  if (refFrArr_top[img_block_y+j][img->block_x+k] == -1)
                  {
                    img->fw_refFrArr_top[img_block_y + j][img->block_x + k] = -1;
                    img->bw_refFrArr_top[img_block_y + j][img->block_x + k] = -1;
                  }
                  else
                  {
                    img->fw_refFrArr_top[img_block_y + j][img->block_x + k] = refFrArr_top[img_block_y+j][img->block_x+k];
                    img->bw_refFrArr_top[img_block_y + j][img->block_x + k] = 0;
                  }
                }
                else
                {
                  img_block_y   = (img->block_y-4)/2;
                  if (refFrArr_bot[img_block_y+j][img->block_x+k] == -1)
                  {
                    img->fw_refFrArr_bot[img_block_y + j][img->block_x + k] = -1;
                    img->bw_refFrArr_bot[img_block_y + j][img->block_x + k] = -1;
                  }
                  else
                  {
                    img->fw_refFrArr_bot[img_block_y + j][img->block_x + k] = refFrArr_bot[img_block_y+j][img->block_x+k];
                    img->bw_refFrArr_bot[img_block_y + j][img->block_x + k] = 0;
                  }
                }
              }
              else
              {
                if (refFrArr[img->block_y+j][img->block_x+k] == -1)
                {
                  img->fw_refFrArr[img->block_y + j][img->block_x + k] = -1;
                  img->bw_refFrArr[img->block_y + j][img->block_x + k] = -1;
                }
                else
                {
                  if (img->structure == TOP_FIELD)
                    img->fw_refFrArr[img->block_y + j][img->block_x + k] = refFrArr[img->block_y+j][img->block_x+k] + 0;    // PLUS2, Krit, 7/06 (used to be + 1)
                  else
                    img->fw_refFrArr[img->block_y + j][img->block_x + k] = refFrArr[img->block_y+j][img->block_x+k];
                  img->bw_refFrArr[img->block_y + j][img->block_x + k] = 0;
                }
              }
            }
          }
        }
      }
    }
  } 

  //  If multiple ref. frames, read reference frame for the MB *********************************
  if(img->num_ref_idx_l0_active>1) 
  {
    flag_mode = ( img->num_ref_idx_l0_active == 2 ? 1 : 0);

    currSE.type = SE_REFFRAME;
    if (bframe)                                               dP = &(currSlice->partArr[partMap[SE_BFRAME]]);
    else                                                      dP = &(currSlice->partArr[partMap[SE_REFFRAME]]);

    if (active_pps->entropy_coding_mode == UVLC || dP->bitstream->ei_flag)   currSE.mapping = linfo_ue;
    else                                                      currSE.reading = readRefFrameFromBuffer_CABAC;
    
    for (j0=0; j0<4; j0+=step_v0)
    {
      for (i0=0; i0<4; i0+=step_h0)
      {
        k=2*(j0/2)+(i0/2);
        if ((currMB->b8pdir[k]==0 || currMB->b8pdir[k]==2) && currMB->b8mode[k]!=0)
        {
          TRACE_STRING("ref_idx_l0");

          img->subblock_x = i0;
          img->subblock_y = j0;
          
          if (!IS_P8x8 (currMB) || bframe || (!bframe && !img->allrefzero))
          {
            currSE.context = BType2CtxRef (currMB->b8mode[k]);
            if( (active_pps->entropy_coding_mode == UVLC || dP->bitstream->ei_flag) && flag_mode )
            {
              currSE.len = 1;
              readSyntaxElement_FLC(&currSE, dP->bitstream);
              currSE.value1 = 1 - currSE.value1;
            }
            else
            {
              dP->readSyntaxElement (&currSE,img,inp,dP);
            }
            refframe = currSE.value1;
            
          }
          else
          {
            refframe = 0;
          }
          
          if (bframe && refframe>img->buf_cycle)    // img->buf_cycle should be correct for field MBs now
          {
            set_ec_flag(SE_REFFRAME);
            refframe = 1;
            
          }
          
          for (j=j0; j<j0+step_v0;j++)
            for (i=i0; i<i0+step_h0;i++)
            {
              dec_picture->ref_idx[LIST_0][img->block_x + i][img->block_y + j] = refframe;
            }
          
        }
      }
    }
  }
  else
  {
    for (j0=0; j0<4; j0+=step_v0)
    {
      for (i0=0; i0<4; i0+=step_h0)
      {
        k=2*(j0/2)+(i0/2);
        if ((currMB->b8pdir[k]==0 || currMB->b8pdir[k]==2) && currMB->b8mode[k]!=0)
        {
          for (j=j0; j<j0+step_v0;j++)
            for (i=i0; i<i0+step_h0;i++)
            {
              dec_picture->ref_idx[LIST_0][img->block_x + i][img->block_y + j] = 0;
            }
        }
      }
    }
  }
  
  //  If backward multiple ref. frames, read backward reference frame for the MB *********************************
  if(img->num_ref_idx_l1_active>1)
  {
    flag_mode = ( img->num_ref_idx_l1_active == 2 ? 1 : 0);

    currSE.type = SE_BFRAME;
    dP = &(currSlice->partArr[partMap[SE_BFRAME]]);
    if (active_pps->entropy_coding_mode == UVLC || dP->bitstream->ei_flag)   currSE.mapping = linfo_ue;
    else                                                      currSE.reading = readBwdRefFrameFromBuffer_CABAC;
    
    for (j0=0; j0<4; j0+=step_v0)
    {
      for (i0=0; i0<4; i0+=step_h0)
      {
        k=2*(j0/2)+(i0/2);
        if ((currMB->b8pdir[k]==1 || currMB->b8pdir[k]==2) && currMB->b8mode[k]!=0)
        {
          TRACE_STRING("ref_idx_l1");

          img->subblock_x = i0;
          img->subblock_y = j0;
          
          currSE.context = BType2CtxRef (currMB->b8mode[k]);
          if( (active_pps->entropy_coding_mode == UVLC || dP->bitstream->ei_flag) && flag_mode )
          {
            currSE.len = 1;
            readSyntaxElement_FLC(&currSE, dP->bitstream);
            currSE.value1 = 1-currSE.value1;
          }
          else
          {
            dP->readSyntaxElement (&currSE,img,inp,dP);
          }
          refframe = currSE.value1;

          for (j=j0; j<j0+step_v0;j++)
          {
            for (i=i0; i<i0+step_h0;i++)
            {
              dec_picture->ref_idx[LIST_1][img->block_x + i][img->block_y + j] = refframe;
            }
          }
        }
      }
    }
  }
  else
  {
    for (j0=0; j0<4; j0+=step_v0)
    {
      for (i0=0; i0<4; i0+=step_h0)
      {
        k=2*(j0/2)+(i0/2);
        if ((currMB->b8pdir[k]==1 || currMB->b8pdir[k]==2) && currMB->b8mode[k]!=0)
        {
          for (j=j0; j<j0+step_v0;j++)
            for (i=i0; i<i0+step_h0;i++)
            {
              dec_picture->ref_idx[LIST_1][img->block_x + i][img->block_y + j] = 0;
            }
        }
      }
    }
  }

  //=====  READ FORWARD MOTION VECTORS =====
  currSE.type = SE_MVD;
  if (bframe)   dP = &(currSlice->partArr[partMap[SE_BFRAME]]);
  else          dP = &(currSlice->partArr[partMap[SE_MVD]]);

  if (active_pps->entropy_coding_mode == UVLC || dP->bitstream->ei_flag) currSE.mapping = linfo_se;
  else if (bframe)                                        currSE.reading = readBiMVD2Buffer_CABAC;
  else                                                    currSE.reading = readMVDFromBuffer_CABAC;

  for (j0=0; j0<4; j0+=step_v0)
    for (i0=0; i0<4; i0+=step_h0)
    {
      k=2*(j0/2)+(i0/2);

      if ((currMB->b8pdir[k]==0 || currMB->b8pdir[k]==2) && (currMB->b8mode[k] !=0))//has forward vector
      {
        mv_mode  = currMB->b8mode[k];
        step_h   = BLOCK_STEP [mv_mode][0];
        step_v   = BLOCK_STEP [mv_mode][1];
        
        refframe = dec_picture->ref_idx[LIST_0][img->block_x+i0][img->block_y+j0];
        
        for (j=j0; j<j0+step_v0; j+=step_v)
        {
          for (i=i0; i<i0+step_h0; i+=step_h)
          {
            j4 = img->block_y+j;
            i4 = img->block_x+i;
            
            // first make mv-prediction
            SetMotionVectorPredictor (img, pmv, pmv+1, refframe, LIST_0, dec_picture->ref_idx, dec_picture->mv, i, j, 4*step_h, 4*step_v);

            for (k=0; k < 2; k++) 
            {
              TRACE_STRING("mvd_l0");

              img->subblock_x = i; // position used for context determination
              img->subblock_y = j; // position used for context determination
              currSE.value2 = (!bframe ? k : 2*k); // identifies the component; only used for context determination
              dP->readSyntaxElement(&currSE,img,inp,dP);
              curr_mvd = currSE.value1; 
              
              vec=curr_mvd+pmv[k];           /* find motion vector */
              
              for(ii=0;ii<step_h;ii++)
              {
                for(jj=0;jj<step_v;jj++)
                {
                  dec_picture->mv  [LIST_0][i4+ii][j4+jj][k] = vec;
                  currMB->mvd      [LIST_0][j+jj] [i+ii] [k] = curr_mvd;
                }
              }
            }
          }
        }
      }
      else if (currMB->b8mode[k=2*(j0/2)+(i0/2)]==0)      
      {  
        if (!img->direct_type)
        {
          
          ref=currMB->mb_field ? (img->mb_y%2 ? refFrArr_bot[(img->block_y-4)/2+j0][img->block_x+i0]
            :refFrArr_top[img->block_y/2+j0][img->block_x+i0] ):refFrArr[img->block_y+j0][img->block_x+i0];          
          
          if (ref==-1)
          {
            img_block_y = (img->current_mb_nr%2) ? (img->block_y-4)/2:img->block_y/2;
            for (j=j0; j<j0+step_v0; j++)
              for (i=i0; i<i0+step_h0; i++)
              {            
                if(img->mb_frame_field_flag&&currMB->mb_field && (img->current_mb_nr%2 == 0))
                {
                  img->fw_refFrArr_top[img_block_y+j][img->block_x+i]=-1;
                  img->bw_refFrArr_top[img_block_y+j][img->block_x+i]=-1;
                }
                else if(img->mb_frame_field_flag&&currMB->mb_field && (img->current_mb_nr%2 != 0))
                {
                  img->fw_refFrArr_bot[img_block_y+j][img->block_x+i]=-1;
                  img->bw_refFrArr_bot[img_block_y+j][img->block_x+i]=-1;
                }
                else if(img->mb_frame_field_flag)
                {
                  img->fw_refFrArr_frm[img->block_y+j][img->block_x+i]=-1;
                  img->bw_refFrArr_frm[img->block_y+j][img->block_x+i]=-1;
                }
                else
                {
                  img->fw_refFrArr[img->block_y+j][img->block_x+i]=-1; 
                  img->bw_refFrArr[img->block_y+j][img->block_x+i]=-1;
                }            
                j4 = img->block_y+j;
                i4 = img->block_x+i;            
                for (ii=0; ii < 2; ii++) 
                {
                  img->fw_mv[i4+BLOCK_SIZE][j4][ii]=0;
                  img->bw_mv[i4+BLOCK_SIZE][j4][ii]=0;
                  if (img->mb_frame_field_flag && currMB->mb_field)
                  {
                    if (img->current_mb_nr%2 == 0)
                    {
                      j4 = img->block_y/2 + j;
                      img->fw_mv_top[i4+BLOCK_SIZE][j4][ii]=0;
                      img->bw_mv_top[i4+BLOCK_SIZE][j4][ii]=0;
                      img->fw_refFrArr_top[j4][i4]=-1;
                      img->bw_refFrArr_top[j4][i4]=-1;
                    }
                    else
                    {
                      j4 = (img->block_y-4)/2 + j;
                      img->fw_mv_bot[i4+BLOCK_SIZE][j4][ii]=0;
                      img->bw_mv_bot[i4+BLOCK_SIZE][j4][ii]=0;
                      img->fw_refFrArr_bot[j4][i4]=-1;
                      img->bw_refFrArr_bot[j4][i4]=-1;
                    }
                  }
                  if (img->mb_frame_field_flag && currMB->mb_field == 0)
                  {
                    img->fw_mv_frm[i4+BLOCK_SIZE][j4][ii]=0;
                    img->bw_mv_frm[i4+BLOCK_SIZE][j4][ii]=0;
                  }
                }
              }
          }
          else 
          {        
            for (j=j0; j<j0+step_v0; j++)
              for (i=i0; i<i0+step_h0; i++)
              { 
                ref=currMB->mb_field ? (img->mb_y%2 ? refFrArr_bot[(img->block_y-4)/2+j][img->block_x+i]
                  :refFrArr_top[img->block_y/2+j][img->block_x+i] ):refFrArr[img->block_y+j][img->block_x+i];
                img_block_y = (img->current_mb_nr%2) ? (img->block_y-4)/2:img->block_y/2;
                {
                  frame_no_next_P =img->imgtr_next_P+((img->structure==BOTTOM_FIELD));
                  frame_no_B = (img->structure==TOP_FIELD || img->structure==BOTTOM_FIELD) ? img->tr_fld : 2*img->tr_frm;
                  
                  delta_P = (img->imgtr_next_P - img->imgtr_last_P);
                  if((img->structure==TOP_FIELD)) // top field
                  {
                    iTRp = delta_P*(ref/2+1)-(ref+1)%2;
                  }
                  else if((img->structure==BOTTOM_FIELD)) // bot field
                  {
                    iTRp = 1+delta_P*((ref+1)/2)-ref%2;
                  }
                  else  // frame
                  {
                    iTRp = (ref+1)*delta_P;
                    if(img->mb_frame_field_flag && currMB->mb_field)
                    {
                      if(img->mb_y%2)
                        iTRp = 1+delta_P*((ref+1)/2)-ref%2;
                      else
                        iTRp = delta_P*(ref/2+1)-(ref+1)%2;
                    }
                    
                  }
                  
                  iTRb = iTRp - (frame_no_next_P - frame_no_B);
                  mv_scale = (iTRb * 256)/ iTRp;  //! Note that this could be precomputed at the frame/slice level. 
                  //! In such case, an array is needed for each different reference.
                  if (img->structure == TOP_FIELD)
                  {
                    img->fw_refFrArr[img->block_y+j][img->block_x+i]=ref + 0; // PLUS2, Krit, 7/06 (used to be + 1)
                    img->bw_refFrArr[img->block_y+j][img->block_x+i]=0;
                  }
                  else   
                  {
                    if(img->mb_frame_field_flag&&currMB->mb_field && (img->current_mb_nr%2 == 0))
                    {
                      img->fw_refFrArr_top[img_block_y+j][img->block_x+i]=ref;
                      img->bw_refFrArr_top[img_block_y+j][img->block_x+i]=0;
                    }
                    else if(img->mb_frame_field_flag&&currMB->mb_field && (img->current_mb_nr%2 != 0))
                    {
                      img->fw_refFrArr_bot[img_block_y+j][img->block_x+i]=ref;
                      img->bw_refFrArr_bot[img_block_y+j][img->block_x+i]=0;
                    }
                    else if(img->mb_frame_field_flag)
                    {
                      img->fw_refFrArr_frm[img->block_y+j][img->block_x+i]=ref;
                      img->bw_refFrArr_frm[img->block_y+j][img->block_x+i]=0;
                    }
                    else
                    {
                      img->fw_refFrArr[img->block_y+j][img->block_x+i]=ref + 0; // PLUS2, Krit, 7/06 (used to be + 1)
                      img->bw_refFrArr[img->block_y+j][img->block_x+i]=0;
                    }
                  }
                  j4 = img->block_y+j;
                  i4 = img->block_x+i;
                  for (ii=0; ii < 2; ii++) 
                  {              
                    img->fw_mv[i4+BLOCK_SIZE][j4][ii]= (mv_scale * img->mv[i4+BLOCK_SIZE][j4][ii] + 128)>>8;
                    img->bw_mv[i4+BLOCK_SIZE][j4][ii]= ((mv_scale - 256) * img->mv[i4+BLOCK_SIZE][j4][ii] + 128)>>8;
                    if (img->mb_frame_field_flag && currMB->mb_field)
                    {
                      if (img->current_mb_nr%2 == 0)
                      {
                        j4 = img->block_y/2 + j;
                        img->fw_mv_top[i4+BLOCK_SIZE][j4][ii]= (mv_scale * img->mv_top[i4+BLOCK_SIZE][j4][ii] + 128)>>8;
                        img->bw_mv_top[i4+BLOCK_SIZE][j4][ii]=((mv_scale - 256) * img->mv_top[i4+BLOCK_SIZE][j4][ii] + 128)>>8;
                      }
                      else
                      {
                        j4 = (img->block_y-4)/2 + j;
                        img->fw_mv_bot[i4+BLOCK_SIZE][j4][ii]= (mv_scale * img->mv_bot[i4+BLOCK_SIZE][j4][ii] + 128)>>8;  // TBD -- fix it to use bot mv 
                        img->bw_mv_bot[i4+BLOCK_SIZE][j4][ii]=((mv_scale - 256) * img->mv_bot[i4+BLOCK_SIZE][j4][ii] + 128)>>8; // TBD -- fix it to use bot mv 
                      }
                    }
                    if (img->mb_frame_field_flag && currMB->mb_field == 0)
                    {
                      img->fw_mv_frm[i4+BLOCK_SIZE][j4][ii]= (mv_scale * img->mv[i4+BLOCK_SIZE][j4][ii] + 128)>>8;
                      img->bw_mv_frm[i4+BLOCK_SIZE][j4][ii]=((mv_scale - 256) * img->mv[i4+BLOCK_SIZE][j4][ii] + 128)>>8;
                      
                    }
                  }
                } 
              }
          }  
      } 
    }
  }
  
  //=====  READ BACKWARD MOTION VECTORS =====
  currSE.type = SE_MVD;
  dP          = &(currSlice->partArr[partMap[SE_BFRAME]]);

  if (active_pps->entropy_coding_mode == UVLC || dP->bitstream->ei_flag) currSE.mapping = linfo_se;
  else                                                    currSE.reading = readBiMVD2Buffer_CABAC;

  for (j0=0; j0<4; j0+=step_v0)
  {
    for (i0=0; i0<4; i0+=step_h0)
    {
      k=2*(j0/2)+(i0/2);
      if ((currMB->b8pdir[k]==1 || currMB->b8pdir[k]==2) && (currMB->b8mode[k]!=0))//has backward vector
      {
        mv_mode  = currMB->b8mode[k];
        step_h   = BLOCK_STEP [mv_mode][0];
        step_v   = BLOCK_STEP [mv_mode][1];
        
        refframe = dec_picture->ref_idx[LIST_1][img->block_x+i0][img->block_y+j0];

        // KS: Hmm, strange, didn't find this in the document...
        /*
        use_scaled_mv = 0;
        if(currMB->b8pdir[k]==2)
        {
          refframe_l0 = dec_picture->ref_idx[LIST_0][img->block_x+i0][img->block_y+j0];
          current_tr  = (img->structure==TOP_FIELD || img->structure==BOTTOM_FIELD)?img->tr_fld:2*img->tr_frm;
          if((current_tr >= img->imgtr_next_P) && (current_tr >= img->imgtr_last_P))
          {
            use_scaled_mv = 1;
            mv_scale = ((refframe+1)*256)/(refframe_l0+1);
          }
        }*/

        for (j=j0; j<j0+step_v0; j+=step_v)
        {
          for (i=i0; i<i0+step_h0; i+=step_h)
          {
            j4 = img->block_y+j;
            i4 = img->block_x+i;
            
            // first make mv-prediction
            /*
            if(use_scaled_mv)
            {
              pmv[0] = (mv_scale*dec_picture->mv[LIST_0][i4][j4][0]+128)>>8;
              pmv[1] = (mv_scale*dec_picture->mv[LIST_0][i4][j4][1]+128)>>8;
            }
            else*/
            {
              SetMotionVectorPredictor (img, pmv, pmv+1, refframe, LIST_1, dec_picture->ref_idx, dec_picture->mv, i, j, 4*step_h, 4*step_v);
            }
            
            for (k=0; k < 2; k++) 
            {
              TRACE_STRING("mvd_l1");
              
              img->subblock_x = i; // position used for context determination
              img->subblock_y = j; // position used for context determination
              currSE.value2   = 2*k+1; // identifies the component; only used for context determination
              dP->readSyntaxElement(&currSE,img,inp,dP);
              curr_mvd = currSE.value1; 
              
              vec=curr_mvd+pmv[k];           /* find motion vector */

              for(ii=0;ii<step_h;ii++)
              {
                for(jj=0;jj<step_v;jj++)
                {
                  dec_picture->mv  [LIST_1][i4+ii][j4+jj][k] = vec;
                  currMB->mvd      [LIST_1][j+jj] [i+ii] [k] = curr_mvd;
                }
              }
            }
          }
        }
      }
    }
  }
}



/*!
 ************************************************************************
 * \brief
 *    Get the Prediction from the Neighboring BLocks for Number of Nonzero Coefficients 
 *    
 *    Luma Blocks
 ************************************************************************
 */
int predict_nnz(struct img_par *img, int i,int j)
{
  PixelPos pix;

  int pred_nnz = 0;
  int cnt      = 0;
  int mb_nr    = img->current_mb_nr;

  // left block
  getLuma4x4Neighbour(mb_nr, i, j, -1, 0, &pix);
  if (pix.available)
  {
    pred_nnz = img->nz_coeff [pix.mb_addr ][pix.x][pix.y];
    cnt++;
  }

  // top block
  getLuma4x4Neighbour(mb_nr, i, j, 0, -1, &pix);
  if (pix.available)
  {
    pred_nnz += img->nz_coeff [pix.mb_addr ][pix.x][pix.y];
    cnt++;
  }

  if (cnt==2)
  {
    pred_nnz++;
    pred_nnz/=cnt; 
  }

  return pred_nnz;
}
/*!
 ************************************************************************
 * \brief
 *    Get the Prediction from the Neighboring BLocks for Number of Nonzero Coefficients 
 *    
 *    Chroma Blocks   
 ************************************************************************
 */
int predict_nnz_chroma(struct img_par *img, int i,int j)
{
  PixelPos pix;

  int pred_nnz = 0;
  int cnt      =0;
  int mb_nr    = img->current_mb_nr;

  // left block
  getChroma4x4Neighbour(mb_nr, i%2, j-4, -1, 0, &pix);
  if (pix.available)
  {
    pred_nnz = img->nz_coeff [pix.mb_addr ][2 * (i/2) + pix.x][4 + pix.y];
    cnt++;
  }
  
  // top block
  getChroma4x4Neighbour(mb_nr, i%2, j-4, 0, -1, &pix);
  if (pix.available)
  {
    pred_nnz += img->nz_coeff [pix.mb_addr ][2 * (i/2) + pix.x][4 + pix.y];
    cnt++;
  }

  if (cnt==2)
  {
    pred_nnz++;
    pred_nnz/=cnt; 
  }

  return pred_nnz;
}


/*!
 ************************************************************************
 * \brief
 *    Reads coeff of an 4x4 block (CAVLC)
 *
 * \author
 *    Karl Lillevold <karll@real.com>
 *    contributions by James Au <james@ubvideo.com>
 ************************************************************************
 */


void readCoeff4x4_CAVLC (struct img_par *img,struct inp_par *inp,
                        int block_type, 
                        int i, int j, int levarr[16], int runarr[16],
                        int *number_coefficients)
{
  int mb_nr = img->current_mb_nr;
  Macroblock *currMB = &img->mb_data[mb_nr];
  SyntaxElement currSE;
  Slice *currSlice = img->currentSlice;
  DataPartition *dP;
  int *partMap = assignSE2partition[currSlice->dp_mode];


  int k, code, vlcnum;
  int numcoeff, numtrailingones, numcoeff_vlc;
  int level_two_or_higher;
  int numones, totzeros, level, cdc=0, cac=0;
  int zerosleft, ntr, dptype = 0;
  int max_coeff_num = 0, nnz;
  char type[15];
  int incVlc[] = {0,3,6,12,24,48,32768};    // maximum vlc = 6

  numcoeff = 0;

  switch (block_type)
  {
  case LUMA:
    max_coeff_num = 16;
    sprintf(type, "%s", "Luma");
    if (IS_INTRA (currMB))
    {
      dptype = SE_LUM_AC_INTRA;
    }
    else
    {
      dptype = SE_LUM_AC_INTER;
    }
    break;
  case LUMA_INTRA16x16DC:
    max_coeff_num = 16;
    sprintf(type, "%s", "Lum16DC");
    dptype = SE_LUM_DC_INTRA;
    break;
  case LUMA_INTRA16x16AC:
    max_coeff_num = 15;
    sprintf(type, "%s", "Lum16AC");
    dptype = SE_LUM_AC_INTRA;
    break;

  case CHROMA_DC:
    max_coeff_num = 4;
    cdc = 1;

    sprintf(type, "%s", "ChrDC");
    if (IS_INTRA (currMB))
    {
      dptype = SE_CHR_DC_INTRA;
    }
    else
    {
      dptype = SE_CHR_DC_INTER;
    }
    break;
  case CHROMA_AC:
    max_coeff_num = 15;
    cac = 1;
    sprintf(type, "%s", "ChrAC");
    if (IS_INTRA (currMB))
    {
      dptype = SE_CHR_AC_INTRA;
    }
    else
    {
      dptype = SE_CHR_AC_INTER;
    }
    break;
  default:
    error ("readCoeff4x4_CAVLC: invalid block type", 600);
    break;
  }

  if(img->type == B_SLICE)
  {
    dptype = SE_BFRAME;
  }

  currSE.type = dptype;
  dP = &(currSlice->partArr[partMap[dptype]]);

  img->nz_coeff[img->current_mb_nr][i][j] = 0;


  if (!cdc)
  {
    // luma or chroma AC
    if (!cac)
    {
      nnz = predict_nnz(img, i, j);
    }
    else
    {
      nnz = predict_nnz_chroma(img, i, j);
    }

    if (nnz < 2)
    {
      numcoeff_vlc = 0;
    }
    else if (nnz < 4)
    {
      numcoeff_vlc = 1;
    }
    else if (nnz < 8)
    {
      numcoeff_vlc = 2;
    }
    else //
    {
      numcoeff_vlc = 3;
    }

    currSE.value1 = numcoeff_vlc;

    readSyntaxElement_NumCoeffTrailingOnes(&currSE, dP, type);

    numcoeff =  currSE.value1;
    numtrailingones =  currSE.value2;

    img->nz_coeff[img->current_mb_nr][i][j] = numcoeff;
  }
  else
  {
    // chroma DC
    readSyntaxElement_NumCoeffTrailingOnesChromaDC(&currSE, dP);

    numcoeff =  currSE.value1;
    numtrailingones =  currSE.value2;
  }


  for (k = 0; k < max_coeff_num; k++)
  {
    levarr[k] = 0;
    runarr[k] = 0;
  }

  numones = numtrailingones;
  *number_coefficients = numcoeff;

  if (numcoeff)
  {
    if (numtrailingones)
    {

      currSE.len = numtrailingones;

#if TRACE
      snprintf(currSE.tracestring, 
        TRACESTRING_SIZE, "%s trailing ones sign (%d,%d)", type, i, j);
#endif

      readSyntaxElement_FLC (&currSE, dP->bitstream);

      code = currSE.inf;
      ntr = numtrailingones;
      for (k = numcoeff-1; k > numcoeff-1-numtrailingones; k--)
      {
        ntr --;
        if ((code>>ntr)&1)
          levarr[k] = -1;
        else
          levarr[k] = 1;
      }
    }

    // decode levels
    level_two_or_higher = 1;
    if (numcoeff > 3 && numtrailingones == 3)
      level_two_or_higher = 0;

      if (numcoeff > 10 && numtrailingones < 3)
          vlcnum = 1;
      else
          vlcnum = 0;

    for (k = numcoeff - 1 - numtrailingones; k >= 0; k--)
    {

#if TRACE
      snprintf(currSE.tracestring, 
        TRACESTRING_SIZE, "%s lev (%d,%d) k=%d vlc=%d ", type,
          i, j, k, vlcnum);
#endif

      if (vlcnum == 0)
          readSyntaxElement_Level_VLC0(&currSE, dP);
      else
          readSyntaxElement_Level_VLCN(&currSE, vlcnum, dP);

      if (level_two_or_higher)
      {
          if (currSE.inf > 0)
          currSE.inf ++;
          else
          currSE.inf --;
          level_two_or_higher = 0;
      }

      level = levarr[k] = currSE.inf;
      if (abs(level) == 1)
        numones ++;

      // update VLC table
      if (abs(level)>incVlc[vlcnum])
        vlcnum++;

      if (k == numcoeff - 1 - numtrailingones && abs(level)>3)
        vlcnum = 2;

    }
    
    if (numcoeff < max_coeff_num)
    {
      // decode total run
      vlcnum = numcoeff-1;
      currSE.value1 = vlcnum;

#if TRACE
      snprintf(currSE.tracestring, 
        TRACESTRING_SIZE, "%s totalrun (%d,%d) vlc=%d ", type, i,j, vlcnum);
#endif
      if (cdc)
        readSyntaxElement_TotalZerosChromaDC(&currSE, dP);
      else
        readSyntaxElement_TotalZeros(&currSE, dP);

      totzeros = currSE.value1;
    }
    else
    {
      totzeros = 0;
    }

    // decode run before each coefficient
    zerosleft = totzeros;
    i = numcoeff-1;
    if (zerosleft > 0 && i > 0)
    {
      do 
      {
        // select VLC for runbefore
        vlcnum = zerosleft - 1;
        if (vlcnum > RUNBEFORE_NUM-1)
          vlcnum = RUNBEFORE_NUM-1;

        currSE.value1 = vlcnum;
#if TRACE
        snprintf(currSE.tracestring, 
          TRACESTRING_SIZE, "%s run (%d,%d) k=%d vlc=%d ",
            type, i, j, i, vlcnum);
#endif

        readSyntaxElement_Run(&currSE, dP);
        runarr[i] = currSE.value1;

        zerosleft -= runarr[i];
        i --;
      } while (zerosleft != 0 && i != 0);
    }
    runarr[i] = zerosleft;

  } // if numcoeff
}



/*!
 ************************************************************************
 * \brief
 *    Get coded block pattern and coefficients (run/level)
 *    from the NAL
 ************************************************************************
 */
void readCBPandCoeffsFromNAL(struct img_par *img,struct inp_par *inp)
{
  int i,j,k;
  int level;
  int mb_nr = img->current_mb_nr;
  int ii,jj;
  int i1,j1, m2,jg2;
  Macroblock *currMB = &img->mb_data[mb_nr];
  int cbp;
  SyntaxElement currSE;
  Slice *currSlice = img->currentSlice;
  DataPartition *dP;
  int *partMap = assignSE2partition[currSlice->dp_mode];
  int iii,jjj;
  int coef_ctr, i0, j0, b8;
  int ll;
  int block_x,block_y;
  int start_scan;
  int uv;


  int run, len;
  int levarr[16], runarr[16], numcoeff;

  int qp_per    = (img->qp-MIN_QP)/6;
  int qp_rem    = (img->qp-MIN_QP)%6;
  int qp_per_uv = QP_SCALE_CR[img->qp-MIN_QP]/6;
  int qp_rem_uv = QP_SCALE_CR[img->qp-MIN_QP]%6;
  int smb       = ((img->type==SP_SLICE) && IS_INTER (currMB)) || (img->type == SI_SLICE && currMB->mb_type == SI4MB);

  // read CBP if not new intra mode
  if (!IS_NEWINTRA (currMB))
  {
    if (IS_OLDINTRA (currMB) || currMB->mb_type == SI4MB )   currSE.type = SE_CBP_INTRA;
    else                        currSE.type = SE_CBP_INTER;

    if(img->type == B_SLICE)  dP = &(currSlice->partArr[partMap[SE_BFRAME]]);
    else                    dP = &(currSlice->partArr[partMap[currSE.type]]);
    
    if (active_pps->entropy_coding_mode == UVLC || dP->bitstream->ei_flag)
    {
      if (IS_OLDINTRA (currMB) || currMB->mb_type == SI4MB)  currSE.mapping = linfo_cbp_intra;
      else                       currSE.mapping = linfo_cbp_inter;
    }
    else
    {
      currSE.reading = readCBPFromBuffer_CABAC;
    }

    TRACE_STRING("coded_block_pattern");

    dP->readSyntaxElement(&currSE,img,inp,dP);
    currMB->cbp = cbp = currSE.value1;
    // Delta quant only if nonzero coeffs
    if (cbp !=0)
    {
      if (IS_INTER (currMB))  currSE.type = SE_DELTA_QUANT_INTER;
      else                    currSE.type = SE_DELTA_QUANT_INTRA;

      if(img->type == B_SLICE)  dP = &(currSlice->partArr[partMap[SE_BFRAME]]);
      else                    dP = &(currSlice->partArr[partMap[currSE.type]]);
      
      if (active_pps->entropy_coding_mode == UVLC || dP->bitstream->ei_flag)
      {
        currSE.mapping = linfo_se;
      }
      else
                currSE.reading= readDquant_FromBuffer_CABAC; //gabi

      TRACE_STRING("mb_qp_delta");

      dP->readSyntaxElement(&currSE,img,inp,dP);
      currMB->delta_quant = currSE.value1;
      img->qp= (img->qp-MIN_QP+currMB->delta_quant+(MAX_QP-MIN_QP+1))%(MAX_QP-MIN_QP+1)+MIN_QP;
    }
  }
  else
  {
    cbp = currMB->cbp;
  }

  for (i=0;i<BLOCK_SIZE;i++)
    for (j=0;j<BLOCK_SIZE;j++)
      for(iii=0;iii<BLOCK_SIZE;iii++)
        for(jjj=0;jjj<BLOCK_SIZE;jjj++)
          img->cof[i][j][iii][jjj]=0;// reset luma coeffs


  if (IS_NEWINTRA (currMB)) // read DC coeffs for new intra modes
  {
    currSE.type = SE_DELTA_QUANT_INTRA;

    if(img->type == B_SLICE)  dP = &(currSlice->partArr[partMap[SE_BFRAME]]);
    else                    dP = &(currSlice->partArr[partMap[currSE.type]]);
    
    if (active_pps->entropy_coding_mode == UVLC || dP->bitstream->ei_flag)
    {
      currSE.mapping = linfo_se;
    }
    else
    {
            currSE.reading= readDquant_FromBuffer_CABAC;
      //currSE.reading= readDquant_intra_FromBuffer_CABAC;
    }
#if TRACE
    snprintf(currSE.tracestring, TRACESTRING_SIZE, "Delta quant ");
#endif
    dP->readSyntaxElement(&currSE,img,inp,dP);
    currMB->delta_quant = currSE.value1;
    img->qp= (img->qp-MIN_QP+currMB->delta_quant+(MAX_QP-MIN_QP+1))%(MAX_QP-MIN_QP+1)+MIN_QP;

    for (i=0;i<BLOCK_SIZE;i++)
      for (j=0;j<BLOCK_SIZE;j++)
        img->ipredmode[img->block_x+i][img->block_y+j]=DC_PRED;


    if (active_pps->entropy_coding_mode == UVLC)
    {
      readCoeff4x4_CAVLC(img, inp, LUMA_INTRA16x16DC, 0, 0,
                          levarr, runarr, &numcoeff);

      coef_ctr=-1;
      level = 1;                            // just to get inside the loop
      for(k = 0; k < numcoeff; k++)
      {
        if (levarr[k] != 0)                     // leave if len=1
        {
          coef_ctr=coef_ctr+runarr[k]+1;

          if ((img->structure == FRAME) && (!currMB->mb_field)) 
          {
            i0=SNGL_SCAN[coef_ctr][0];
            j0=SNGL_SCAN[coef_ctr][1];
          }
          else 
          { // Alternate scan for field coding
            i0=FIELD_SCAN[coef_ctr][0];
            j0=FIELD_SCAN[coef_ctr][1];
          }

          img->cof[i0][j0][0][0]=levarr[k];// add new intra DC coeff
        }
      }
    }
    else
    {

      currSE.type = SE_LUM_DC_INTRA;
      if(img->type == B_SLICE)  dP = &(currSlice->partArr[partMap[SE_BFRAME]]);
      else                      dP = &(currSlice->partArr[partMap[currSE.type]]);

      currSE.context      = LUMA_16DC;
      currSE.type         = SE_LUM_DC_INTRA;
      img->is_intra_block = 1;

      if (active_pps->entropy_coding_mode == UVLC || dP->bitstream->ei_flag)
      {
        currSE.mapping = linfo_levrun_inter;
      }
      else
      {
        currSE.reading = readRunLevelFromBuffer_CABAC;
      }



      coef_ctr=-1;
      level = 1;                            // just to get inside the loop
      for(k=0;(k<17) && (level!=0);k++)
      {
#if TRACE
        snprintf(currSE.tracestring, TRACESTRING_SIZE, "DC luma 16x16 ");
#endif
        dP->readSyntaxElement(&currSE,img,inp,dP);
        level = currSE.value1;
        run   = currSE.value2;
        len   = currSE.len;

        if (level != 0)                     // leave if len=1
        {
          coef_ctr=coef_ctr+run+1;

          if ((img->structure == FRAME) && (!currMB->mb_field)) 
          {
            i0=SNGL_SCAN[coef_ctr][0];
            j0=SNGL_SCAN[coef_ctr][1];
          }
          else 
          { // Alternate scan for field coding
            i0=FIELD_SCAN[coef_ctr][0];
            j0=FIELD_SCAN[coef_ctr][1];
          }

          img->cof[i0][j0][0][0]=level;// add new intra DC coeff
        }
      }
    }
    itrans_2(img);// transform new intra DC
  }

  qp_per    = (img->qp-MIN_QP)/6;
  qp_rem    = (img->qp-MIN_QP)%6;
  qp_per_uv = QP_SCALE_CR[img->qp-MIN_QP]/6;
  qp_rem_uv = QP_SCALE_CR[img->qp-MIN_QP]%6;
  currMB->qp = img->qp;

  // luma coefficients
  for (block_y=0; block_y < 4; block_y += 2) /* all modes */
  {
    for (block_x=0; block_x < 4; block_x += 2)
    {

      b8 = 2*(block_y/2) + block_x/2;
      if (active_pps->entropy_coding_mode == UVLC)
      {
        for (j=block_y; j < block_y+2; j++)
        {
          for (i=block_x; i < block_x+2; i++)
          {
            ii = block_x/2; jj = block_y/2;
            b8 = 2*jj+ii;

            if (cbp & (1<<b8))  /* are there any coeff in current block at all */
            {
              if (IS_NEWINTRA(currMB))
              {
                readCoeff4x4_CAVLC(img, inp, LUMA_INTRA16x16AC, i, j,
                                    levarr, runarr, &numcoeff);

                start_scan = 1;
              }
              else
              {
                readCoeff4x4_CAVLC(img, inp, LUMA, i, j,
                                    levarr, runarr, &numcoeff);
                start_scan = 0;
              }

              coef_ctr = start_scan-1;
              for (k = 0; k < numcoeff; k++)
              {
                if (levarr[k] != 0)
                {
                  coef_ctr             += runarr[k]+1;

                  if ((img->structure == FRAME) && (!currMB->mb_field)) 
                  {
                    i0=SNGL_SCAN[coef_ctr][0];
                    j0=SNGL_SCAN[coef_ctr][1];
                  }
                  else { // Alternate scan for field coding
                    i0=FIELD_SCAN[coef_ctr][0];
                    j0=FIELD_SCAN[coef_ctr][1];
                  }
                  currMB->cbp_blk      |= 1 << ((j<<2) + i) ;
                  img->cof[i][j][i0][j0]= levarr[k]*dequant_coef[qp_rem][i0][j0]<<qp_per;
                }
              }
            }
            else
            {
              img->nz_coeff[img->current_mb_nr][i][j] = 0;
            }
          }
        }
      } // VLC
      else
      { 
          b8 = 2*(block_y/2) + block_x/2;
          // CABAC 
          for (j=block_y; j < block_y+2; j++)
          {
            //jj=j/2;
            for (i=block_x; i < block_x+2; i++)
            {
              //ii = i/2;
              //b8 = 2*jj+ii;

              if (IS_NEWINTRA (currMB))   start_scan = 1; // skip DC coeff
              else                        start_scan = 0; // take all coeffs

              img->subblock_x = i; // position for coeff_count ctx
              img->subblock_y = j; // position for coeff_count ctx
              if (cbp & (1<<b8))  // are there any coeff in current block at all
              {
                coef_ctr = start_scan-1;
                level    = 1;
                for(k=start_scan;(k<17) && (level!=0);k++)
                {
                 /*
                  * make distinction between INTRA and INTER coded
                  * luminance coefficients
                  */
                  currSE.context      = (IS_NEWINTRA(currMB) ? LUMA_16AC : LUMA_4x4);
                  currSE.type         = (IS_INTRA(currMB) ?
                                        (k==0 ? SE_LUM_DC_INTRA : SE_LUM_AC_INTRA) :
                                        (k==0 ? SE_LUM_DC_INTER : SE_LUM_AC_INTER));
                  img->is_intra_block = IS_INTRA(currMB);
                  
#if TRACE
                  sprintf(currSE.tracestring, " Luma sng ");
#endif
                  if(img->type == B_SLICE)  dP = &(currSlice->partArr[partMap[SE_BFRAME]]);
                  else                    dP = &(currSlice->partArr[partMap[currSE.type]]);
                  
                  if (active_pps->entropy_coding_mode == UVLC || dP->bitstream->ei_flag)  currSE.mapping = linfo_levrun_inter;
                  else                                                     currSE.reading = readRunLevelFromBuffer_CABAC;
                  
                  dP->readSyntaxElement(&currSE,img,inp,dP);
                  level = currSE.value1;
                  run   = currSE.value2;
                  len   = currSE.len;
                  
                  if (level != 0)    /* leave if len=1 */
                  {
                    coef_ctr             += run+1;

                    if ((img->structure == FRAME) && (!currMB->mb_field)) 
                    {
                      i0=SNGL_SCAN[coef_ctr][0];
                      j0=SNGL_SCAN[coef_ctr][1];
                    }
                    else { // Alternate scan for field coding
                      i0=FIELD_SCAN[coef_ctr][0];
                      j0=FIELD_SCAN[coef_ctr][1];
                    }
                    currMB->cbp_blk      |= 1 << ((j<<2) + i) ;
                    img->cof[i][j][i0][j0]= level*dequant_coef[qp_rem][i0][j0]<<qp_per;
                  }
                }
              }
            }
          }
      } 
    }
  }

  for (j=4;j<6;j++) // reset all chroma coeffs before read
    for (i=0;i<4;i++)
      for (iii=0;iii<4;iii++)
        for (jjj=0;jjj<4;jjj++)
          img->cof[i][j][iii][jjj]=0;

  m2 =img->mb_x*2;
  jg2=img->mb_y*2;

  // chroma 2x2 DC coeff
  if(cbp>15)
  {
    for (ll=0;ll<3;ll+=2)
    {
      for (i=0;i<4;i++)
        img->cofu[i]=0;


      if (active_pps->entropy_coding_mode == UVLC)
      {

        readCoeff4x4_CAVLC(img, inp, CHROMA_DC, 0, 0,
                            levarr, runarr, &numcoeff);
        coef_ctr=-1;
        level=1;
        for(k = 0; k < numcoeff; k++)
        {
          if (levarr[k] != 0)
          {
            currMB->cbp_blk |= 0xf0000 << (ll<<1) ;
            coef_ctr=coef_ctr+runarr[k]+1;
            img->cofu[coef_ctr]=levarr[k];
          }
        }
      }
      else
      {
        coef_ctr=-1;
        level=1;
        for(k=0;(k<5)&&(level!=0);k++)
        {
          currSE.context      = CHROMA_DC;
          currSE.type         = (IS_INTRA(currMB) ? SE_CHR_DC_INTRA : SE_CHR_DC_INTER);
          img->is_intra_block =  IS_INTRA(currMB);
          img->is_v_block     = ll;

#if TRACE
          snprintf(currSE.tracestring, TRACESTRING_SIZE, " 2x2 DC Chroma ");
#endif
          if(img->type == B_SLICE)
            dP = &(currSlice->partArr[partMap[SE_BFRAME]]);
          else
            dP = &(currSlice->partArr[partMap[currSE.type]]);
        
          if (active_pps->entropy_coding_mode == UVLC || dP->bitstream->ei_flag)
            currSE.mapping = linfo_levrun_c2x2;
          else
            currSE.reading = readRunLevelFromBuffer_CABAC;

          dP->readSyntaxElement(&currSE,img,inp,dP);
          level = currSE.value1;
          run = currSE.value2;
          len = currSE.len;
          if (level != 0)
          {
            currMB->cbp_blk |= 0xf0000 << (ll<<1) ;
            coef_ctr=coef_ctr+run+1;
            // Bug: img->cofu has only 4 entries, hence coef_ctr MUST be <4 (which is
            // caught by the assert().  If it is bigger than 4, it starts patching the
            // img->predmode pointer, which leads to bugs later on.
            //
            // This assert() should be left in the code, because it captures a very likely
            // bug early when testing in error prone environments (or when testing NAL
            // functionality).
            assert (coef_ctr < 4);
            img->cofu[coef_ctr]=level;
          }
        }
      }

      if (smb) // check to see if MB type is SPred or SIntra4x4 
      {
         img->cof[0+ll][4][0][0]=img->cofu[0];   img->cof[1+ll][4][0][0]=img->cofu[1];
         img->cof[0+ll][5][0][0]=img->cofu[2];   img->cof[1+ll][5][0][0]=img->cofu[3];
      }
      else
      { 
        for (i=0;i<4;i++)
          img->cofu[i]*=dequant_coef[qp_rem_uv][0][0]<<qp_per_uv;
        img->cof[0+ll][4][0][0]=(img->cofu[0]+img->cofu[1]+img->cofu[2]+img->cofu[3])>>1;
        img->cof[1+ll][4][0][0]=(img->cofu[0]-img->cofu[1]+img->cofu[2]-img->cofu[3])>>1;
        img->cof[0+ll][5][0][0]=(img->cofu[0]+img->cofu[1]-img->cofu[2]-img->cofu[3])>>1;
        img->cof[1+ll][5][0][0]=(img->cofu[0]-img->cofu[1]-img->cofu[2]+img->cofu[3])>>1;
      }
    }
  }

  // chroma AC coeff, all zero fram start_scan
  if (cbp<=31)
    for (j=4; j < 6; j++)
      for (i=0; i < 4; i++)
        img->nz_coeff [img->current_mb_nr ][i][j]=0;


  // chroma AC coeff, all zero fram start_scan
  uv=-1;
  if (cbp>31)
  {
    block_y=4;
    for (block_x=0; block_x < 4; block_x += 2)
    {
      for (j=block_y; j < block_y+2; j++)
      {
        jj=j/2;
        j1=j-4;
        for (i=block_x; i < block_x+2; i++)
        {

          ii=i/2;
          i1=i%2;

          if (active_pps->entropy_coding_mode == UVLC)
          {
            readCoeff4x4_CAVLC(img, inp, CHROMA_AC, i, j,
                                levarr, runarr, &numcoeff);
            coef_ctr=0;
            level=1;
            uv++;
            for(k = 0; k < numcoeff;k++)
            {
              if (levarr[k] != 0)
              {
                currMB->cbp_blk |= 1 << (16 + (j1<<1) + i1 + (block_x<<1) ) ;
                coef_ctr=coef_ctr+runarr[k]+1;

                if ((img->structure == FRAME) && (!currMB->mb_field)) 
                {
                  i0=SNGL_SCAN[coef_ctr][0];
                  j0=SNGL_SCAN[coef_ctr][1];
                }
                else { // Alternate scan for field coding
                  i0=FIELD_SCAN[coef_ctr][0];
                  j0=FIELD_SCAN[coef_ctr][1];
                }
                img->cof[i][j][i0][j0]=levarr[k]*dequant_coef[qp_rem_uv][i0][j0]<<qp_per_uv;
              }
            }
          }

          else
          {
            coef_ctr=0;
            level=1;
            uv++;

            img->subblock_y = j/5;
            img->subblock_x = i%2;

            for(k=0;(k<16)&&(level!=0);k++)
            {
              currSE.context      = CHROMA_AC;
              currSE.type         = (IS_INTRA(currMB) ? SE_CHR_AC_INTRA : SE_CHR_AC_INTER);
              img->is_intra_block =  IS_INTRA(currMB);
              img->is_v_block     = (uv>=4);

#if TRACE
              snprintf(currSE.tracestring, TRACESTRING_SIZE, " AC Chroma ");
#endif
              if(img->type == B_SLICE)
                dP = &(currSlice->partArr[partMap[SE_BFRAME]]);
              else
                dP = &(currSlice->partArr[partMap[currSE.type]]);
            
              if (active_pps->entropy_coding_mode == UVLC || dP->bitstream->ei_flag)
                currSE.mapping = linfo_levrun_inter;
              else
                currSE.reading = readRunLevelFromBuffer_CABAC;
                       
              dP->readSyntaxElement(&currSE,img,inp,dP);
              level = currSE.value1;
              run = currSE.value2;
              len = currSE.len;

              if (level != 0)
              {
                currMB->cbp_blk |= 1 << (16 + (j1<<1) + i1 + (block_x<<1) ) ;
                coef_ctr=coef_ctr+run+1;

                if ((img->structure == FRAME) && (!currMB->mb_field)) 
                {
                  i0=SNGL_SCAN[coef_ctr][0];
                  j0=SNGL_SCAN[coef_ctr][1];
                }
                else { // Alternate scan for field coding
                  i0=FIELD_SCAN[coef_ctr][0];
                  j0=FIELD_SCAN[coef_ctr][1];
                }
                img->cof[i][j][i0][j0]=level*dequant_coef[qp_rem_uv][i0][j0]<<qp_per_uv;
              }
            }
          }
        }
      }
    }
  }
}



/*!
 ************************************************************************
 * \brief
 *    decode one macroblock
 ************************************************************************
 */

int decode_one_macroblock(struct img_par *img,struct inp_par *inp)
{
  int tmp_block[BLOCK_SIZE][BLOCK_SIZE];
  int tmp_blockbw[BLOCK_SIZE][BLOCK_SIZE];
  int i=0,j=0,k,l,ii=0,jj=0,i1=0,j1=0,j4=0,i4=0;
  int uv, hv;
  int vec1_x=0,vec1_y=0,vec2_x=0,vec2_y=0;
  int ioff,joff;
  int block8x8;   // needed for ABT

  int bw_pred, fw_pred, pred, ifx;
  int ii0,jj0,ii1,jj1,if1,jf1,if0,jf0;
  int mv_mul,f1,f2,f3,f4;

  const byte decode_block_scan[16] = {0,1,4,5,2,3,6,7,8,9,12,13,10,11,14,15};

  Macroblock *currMB   = &img->mb_data[img->current_mb_nr];
  int ref_idx, fw_refframe, bw_refframe, mv_mode, pred_dir, intra_prediction; // = currMB->ref_frame;
  int fw_ref_idx, bw_ref_idx;

  int  *** mv_array, ***fw_mv_array, ***bw_mv_array;
  int bframe = (img->type==B_SLICE);
//  byte refP_tr;

  int **fwRefFrArr = img->fw_refFrArr;
  int **bwRefFrArr = img->bw_refFrArr;
  int  ***fw_mv = img->fw_mv;
  int  ***bw_mv = img->bw_mv;
  int  **moving_block_dir = moving_block; 

  int partmode        = (IS_P8x8(currMB)?4:currMB->mb_type);
  int step_h0 = BLOCK_STEP [partmode][0];
  int step_v0 = BLOCK_STEP [partmode][1];

  int frame_no_next_P, frame_no_B, delta_P;

  int ref;

  int iTRb, iTRp;
  int mv_scale;

  int mb_nr             = img->current_mb_nr;
  int mb_width          = img->width/16;
  int mb_available_up;
  int mb_available_left;
  int smb       = ((img->type==SP_SLICE) && IS_INTER (currMB)) || (img->type == SI_SLICE && currMB->mb_type == SI4MB);
  int list_offset;
  int max_y_cr;

  StorablePicture **list;

  int jf, j6,j5 = 0;

  int direct_pdir;

  int curr_mb_field = ((img->mb_frame_field_flag)&&(currMB->mb_field));

  // find out the correct list offsets
  if (curr_mb_field)
  {
    if(mb_nr%2)
      list_offset = 4; // top field mb
    else
      list_offset = 2; // bottom field mb
    max_y_cr = img->height_cr/2-1;
  }
  else
  {
    list_offset = 0;  // no mb aff or frame mb
    max_y_cr = img->height_cr-1;
  }

  if (!img->mb_frame_field_flag)
  {
    for (l=0+list_offset;l<(2+list_offset);l++)
    {
      for(k = 0; k < listXsize[l]; k++)
      {
        listX[l][k]->chroma_vector_adjustment= 0;
        if(img->structure == TOP_FIELD && img->structure != listX[l][k]->structure)
          listX[l][k]->chroma_vector_adjustment = -2;
        if(img->structure == BOTTOM_FIELD && img->structure != listX[l][k]->structure)
          listX[l][k]->chroma_vector_adjustment = 2;
      }
    }
  }
  else
  {
    if (curr_mb_field)
    {
      for (l=0+list_offset;l<(2+list_offset);l++)
      {
        for(k = 0; k < listXsize[l]; k++)
        {
          listX[l][k]->chroma_vector_adjustment= 0;
          if(img->current_mb_nr % 2 == 0 && listX[l][k]->structure == BOTTOM_FIELD)
            listX[l][k]->chroma_vector_adjustment = -2;
          if(img->current_mb_nr % 2 == 1 && listX[l][k]->structure == TOP_FIELD)
            listX[l][k]->chroma_vector_adjustment = 2;
        }
      }
    }
    else
    {
      for (l=0+list_offset;l<(2+list_offset);l++)
      {
        for(k = 0; k < listXsize[l]; k++)
        {
          listX[l][k]->chroma_vector_adjustment= 0;
        }
      }
    }
  }


  if (img->mb_frame_field_flag)
  {
    mb_available_up   = (img->mb_y == 0) ? 0 : 1;
    mb_available_left = (img->mb_x == 0) ? 0 : 1;
  }
  else
  {
    mb_available_up   = (img->mb_y == 0) ? 0 : (img->mb_data[mb_nr].slice_nr == img->mb_data[mb_nr-mb_width].slice_nr);
    mb_available_left = (img->mb_x == 0) ? 0 : (img->mb_data[mb_nr].slice_nr == img->mb_data[mb_nr-1].slice_nr);
  }

  if(img->constrained_intra_pred_flag)
  {
    if (mb_available_up   && (img->intra_block[mb_nr-mb_width][2]==0 || img->intra_block[mb_nr-mb_width][3]==0))
      mb_available_up   = 0;
    if (mb_available_left && (img->intra_block[mb_nr-       1][1]==0 || img->intra_block[mb_nr       -1][3]==0))
      mb_available_left = 0;
  }

  mv_mul=4;
  f1=8;
  f2=7;

  f3=f1*f1;
  f4=f3/2;

  // luma decoding **************************************************

  // get prediction for INTRA_MB_16x16
  if (IS_NEWINTRA (currMB))
  {
    intrapred_luma_16x16(img, currMB->i16mode);
  }

  for (block8x8=0; block8x8<4; block8x8++)
  {
    
#ifdef MBINTLC
    // skipped MB processing for field MB  ***********************************
    if (IS_DIRECT (currMB) && img->mb_frame_field_flag && img->current_mb_nr%2 ==0)      
      decode_skip_Direct_topMB(img,inp);
#endif
    for (k = block8x8*4; k < block8x8*4+4; k ++)
    {
      i = (decode_block_scan[k] & 3);
      j = ((decode_block_scan[k] >> 2) & 3);
      
      ioff=i*4;
      i4=img->block_x+i;
      
      joff=j*4;
      j4=img->block_y+j;

      if (img->mb_frame_field_flag) 
      {
        j5 = img->block_y / 2 + j;
        if (img->current_mb_nr%2)
          j5 -= BLOCK_SIZE / 2;
      }
      mv_mode  = currMB->b8mode[2*(j/2)+(i/2)];
      pred_dir = currMB->b8pdir[2*(j/2)+(i/2)];
      
      assert (pred_dir<=2);

      // PREDICTION
      if (mv_mode==IBLOCK)
      {
        //===== INTRA PREDICTION =====
        if (intrapred(img,ioff,joff,i4,j4)==SEARCH_SYNC)  /* make 4x4 prediction block mpr from given prediction img->mb_mode */
          return SEARCH_SYNC;                   /* bit error */
      }
      else if (!IS_NEWINTRA (currMB))
      {
        if (pred_dir != 2)
        {
          //===== FORWARD/BACKWARD PREDICTION =====
          ref_idx  = dec_picture->ref_idx[LIST_0 + pred_dir][i4][j4];
          mv_array = dec_picture->mv[LIST_0 + pred_dir];
          list     = listX[0+list_offset+ pred_dir];
          
          vec1_x = i4*4*mv_mul + mv_array[i4][j4][0];

          if (!curr_mb_field)
          {
            vec1_y = j4*4*mv_mul + mv_array[i4][j4][1];
          }
          else
          {
            if (mb_nr%2 == 0) 
              vec1_y = (img->block_y * 2 + joff) * mv_mul + mv_array[i4][j4][1];
            else
              vec1_y = ((img->block_y-4) * 2 + joff)* mv_mul + mv_array[i4][j4][1];
          }

          get_block (ref_idx, list, vec1_x, vec1_y, img, tmp_block);

          if (img->apply_weights)
          {
            for(ii=0;ii<BLOCK_SIZE;ii++)
              for(jj=0;jj<BLOCK_SIZE;jj++)  
                img->mpr[ii+ioff][jj+joff] = Clip1(((img->wp_weight[pred_dir][ref_idx][0] *  tmp_block[ii][jj]+ img->wp_round_luma) >>img->luma_log2_weight_denom)  + img->wp_offset[pred_dir][ref_idx][0] );
          }
          else
          {
            for(ii=0;ii<BLOCK_SIZE;ii++)
              for(jj=0;jj<BLOCK_SIZE;jj++)  
                img->mpr[ii+ioff][jj+joff] = tmp_block[ii][jj];
          }
        }
        else
        {
          if (mv_mode != 0)
          {
            //===== BI-DIRECTIONAL PREDICTION =====
            fw_mv_array = dec_picture->mv[LIST_0];
            bw_mv_array = dec_picture->mv[LIST_1];
            
            fw_refframe = dec_picture->ref_idx[LIST_0][i4][j4];
            bw_refframe = dec_picture->ref_idx[LIST_1][i4][j4];
            fw_ref_idx = fw_refframe;
            bw_ref_idx = bw_refframe;
            bw_ref_idx = img->disposable_flag ? bw_ref_idx : 1 - bw_ref_idx;
          }
          else
          {
            //===== DIRECT PREDICTION =====
            fw_mv_array = img->dfMV;
            bw_mv_array = img->dbMV;
            bw_refframe = 0;
            if (img->structure==TOP_FIELD)  bw_refframe = 1;
            
            if (img->direct_type && img->mb_frame_field_flag )
            {
              fwRefFrArr = dec_picture->ref_idx[LIST_0];
              bwRefFrArr = dec_picture->ref_idx[LIST_1];
              fw_mv = dec_picture->mv[LIST_0];
              bw_mv = dec_picture->mv[LIST_1];

              if (img->mb_frame_field_flag)
              {
                if (img->current_mb_nr%2 )
                {
                  moving_block_dir = moving_block_bot; 
                }
                else
                {
                  moving_block_dir = moving_block_top; 
                }
              }
            }

            if (img->direct_type )
            {
              int pic_blockx           = img->block_x;
              int pic_blocky           = (img->mb_frame_field_flag && currMB->mb_field)? img->block_y/2:img->block_y;
              int mb_nr                = img->current_mb_nr;
              int mb_width             = img->width/16;
              int mb_available_up =      (img->mb_y == 0        ) ? 0 : (img->mb_frame_field_flag? 1 :(currMB->slice_nr == img->mb_data[mb_nr-mb_width].slice_nr));
              int mb_available_left =    (img->mb_x == 0        ) ? 0 : (currMB->slice_nr == img->mb_data[mb_nr-1].slice_nr);
              int mb_available_upleft  = (img->mb_x == 0          ||
                img->mb_y == 0        ) ? 0 : (img->mb_frame_field_flag)? 1 :(currMB->slice_nr == img->mb_data[mb_nr-mb_width-1].slice_nr);
              int mb_available_upright = (img->mb_frame_field_flag && img->current_mb_nr%2)?0:(img->mb_x >= mb_width-1 ||
                img->mb_y == 0        ) ? 0 : (img->mb_frame_field_flag)? 1 :(currMB->slice_nr == img->mb_data[mb_nr-mb_width+1].slice_nr);
              
              int fw_rFrameL              = mb_available_left    ? fwRefFrArr[pic_blocky]  [pic_blockx-1]   : -1;
              int fw_rFrameU              = mb_available_up      ? fwRefFrArr[pic_blocky-1][pic_blockx]     : -1;
              int fw_rFrameUL             = mb_available_upleft  ? fwRefFrArr[pic_blocky-1][pic_blockx-1]   : -1;
              int fw_rFrameUR             = mb_available_upright ? fwRefFrArr[pic_blocky-1][pic_blockx+4]   : fw_rFrameUL;  
              
              int bw_rFrameL              = mb_available_left    ? bwRefFrArr[pic_blocky]  [pic_blockx-1]   : -1;
              int bw_rFrameU              = mb_available_up      ? bwRefFrArr[pic_blocky-1][pic_blockx]     : -1;
              int bw_rFrameUL             = mb_available_upleft  ? bwRefFrArr[pic_blocky-1][pic_blockx-1]   : -1;
              int bw_rFrameUR             = mb_available_upright ? bwRefFrArr[pic_blocky-1][pic_blockx+4]   : bw_rFrameUL;
              
              int fw_rFrame,bw_rFrame;
              int pmvfw[2]={0,0},pmvbw[2]={0,0};
              
              
              j6= pic_blocky+j;
              
              if (!fw_rFrameL || !fw_rFrameU || !fw_rFrameUR)
                fw_rFrame=0;
              else
                fw_rFrame=min(fw_rFrameL&15,min(fw_rFrameU&15,fw_rFrameUR&15));
              
              if(img->num_ref_idx_l1_active>1 && (bw_rFrameL==1 || bw_rFrameU==1 || bw_rFrameUR==1))
                bw_rFrame=1;
              else if (!bw_rFrameL || !bw_rFrameU || !bw_rFrameUR)
                bw_rFrame=0;
              else
                bw_rFrame=min(bw_rFrameL&15,min(bw_rFrameU&15,bw_rFrameUR&15));
              
              if (fw_rFrame !=15)
/*                  SetMotionVectorPredictor (img, pmvfw, pmvfw+1, fw_rFrame, fwRefFrArr, fw_mv, 0, 0, 16, 16)*/;  
              if (bw_rFrame !=15)
/*                  SetMotionVectorPredictor (img, pmvbw, pmvbw+1, bw_rFrame, bwRefFrArr, bw_mv, 0, 0, 16, 16)*/;  
                  
              ref=refFrArr[j4][i4];
              if (fw_rFrame !=15)
              {
                if (!fw_rFrame  && !moving_block_dir[j6][i4]) 
                {
                  img->fw_mv[i4+BLOCK_SIZE][j4][0]=img->dfMV[i4+BLOCK_SIZE][j4][0] = 0;
                  img->fw_mv[i4+BLOCK_SIZE][j4][1]=img->dfMV[i4+BLOCK_SIZE][j4][1] = 0;
                  if (img->mb_frame_field_flag)
                  {
                    if (img->current_mb_nr%2 == 0)
                    {
                      img->dfMV_top[i4+BLOCK_SIZE][j5][0]=img->fw_mv_top[i4+BLOCK_SIZE][j5][0]=0;
                      img->dfMV_top[i4+BLOCK_SIZE][j5][1]=img->fw_mv_top[i4+BLOCK_SIZE][j5][1]=0;
                      img->fw_refFrArr_top[j5][i4]=0;
                    }
                    else
                    {
                      img->dfMV_bot[i4+BLOCK_SIZE][j5][0]=img->fw_mv_bot[i4+BLOCK_SIZE][j5][0]=0;
                      img->dfMV_bot[i4+BLOCK_SIZE][j5][1]=img->fw_mv_bot[i4+BLOCK_SIZE][j5][1]=0;
                      img->fw_refFrArr_bot[j5][i4]=0;
                    }
                  } 
                  if (img->structure == TOP_FIELD)
                    fwRefFrArr[j6][i4] = 0;
                  else
                    fwRefFrArr[j6][i4] = 0; 
                }
                else
                {
                  img->fw_mv[i4+BLOCK_SIZE][j4][0]=img->dfMV[i4+BLOCK_SIZE][j4][0] = pmvfw[0];
                  img->fw_mv[i4+BLOCK_SIZE][j4][1]=img->dfMV[i4+BLOCK_SIZE][j4][1] = pmvfw[1];
                  if (img->structure == TOP_FIELD)
                    fwRefFrArr[j6][i4] = fw_rFrame ;
                  else
                    fwRefFrArr[j6][i4] = fw_rFrame; 
                  if (img->mb_frame_field_flag)
                  {
                    if (img->current_mb_nr%2 == 0)
                    {
                      img->dfMV_top[i4+BLOCK_SIZE][j5][0]=img->fw_mv_top[i4+BLOCK_SIZE][j5][0]=pmvfw[0];
                      img->dfMV_top[i4+BLOCK_SIZE][j5][1]=img->fw_mv_top[i4+BLOCK_SIZE][j5][1]=pmvfw[1]/2;
                      img->fw_refFrArr_top[j5][i4]=2*fw_rFrame;
                    }
                    else
                    {
                      img->dfMV_bot[i4+BLOCK_SIZE][j5][0]=img->fw_mv_bot[i4+BLOCK_SIZE][j5][0]=pmvfw[0];
                      img->dfMV_bot[i4+BLOCK_SIZE][j5][1]=img->fw_mv_bot[i4+BLOCK_SIZE][j5][1]=pmvfw[1]/2;
                      img->fw_refFrArr_bot[j5][i4]=2*fw_rFrame;
                    }
                  }
                }
              }
              else
              {
                img->fw_refFrArr[j4][i4]=-1;
                img->fw_mv[i4+BLOCK_SIZE][j4][0]=img->dfMV[i4+BLOCK_SIZE][j4][0] = 0;
                img->fw_mv[i4+BLOCK_SIZE][j4][1]=img->dfMV[i4+BLOCK_SIZE][j4][1] = 0;
                if (img->mb_frame_field_flag)
                {
                  if (img->current_mb_nr%2 == 0)
                  {
                    img->dfMV_top[i4+BLOCK_SIZE][j5][0]=img->fw_mv_top[i4+BLOCK_SIZE][j5][0]=0;
                    img->dfMV_top[i4+BLOCK_SIZE][j5][1]=img->fw_mv_top[i4+BLOCK_SIZE][j5][1]=0;
                    img->fw_refFrArr_top[j5][i4]=-1;
                  }
                  else
                  {
                    img->dfMV_bot[i4+BLOCK_SIZE][j5][0]=img->fw_mv_bot[i4+BLOCK_SIZE][j5][0]=0;
                    img->dfMV_bot[i4+BLOCK_SIZE][j5][1]=img->fw_mv_bot[i4+BLOCK_SIZE][j5][1]=0;
                    img->fw_refFrArr_bot[j5][i4]=-1;
                  }
                }
              }
              
              if (bw_rFrame !=15)
              {
                if  (bw_rFrame==((img->num_ref_idx_l1_active>1)?1:0) && !moving_block_dir[j6][i4])  
                {                  
                  
                  img->bw_mv[i4+BLOCK_SIZE][j4][0]=img->dbMV[i4+BLOCK_SIZE][j4][0] = 0;
                  img->bw_mv[i4+BLOCK_SIZE][j4][1]=img->dbMV[i4+BLOCK_SIZE][j4][1] = 0;
                  bwRefFrArr[j4][i4]=bw_rFrame;
                  if (img->mb_frame_field_flag)
                  {
                    if (img->current_mb_nr%2 == 0)
                    {
                      img->dbMV_top[i4+BLOCK_SIZE][j5][0]=img->bw_mv_top[i4+BLOCK_SIZE][j5][0]=0;
                      img->dbMV_top[i4+BLOCK_SIZE][j5][1]=img->bw_mv_top[i4+BLOCK_SIZE][j5][1]=0;
                      img->bw_refFrArr_top[j5][i4]=2*bw_rFrame;
                    }
                    else
                    {
                      img->dbMV_bot[i4+BLOCK_SIZE][j5][0]=img->bw_mv_bot[i4+BLOCK_SIZE][j5][0]=0;
                      img->dbMV_bot[i4+BLOCK_SIZE][j5][1]=img->bw_mv_bot[i4+BLOCK_SIZE][j5][1]=0;
                      img->bw_refFrArr_bot[j5][i4]=2*bw_rFrame;
                    }
                  }                    
                }
                else
                {
                  img->bw_mv[i4+BLOCK_SIZE][j4][0]=img->dbMV[i4+BLOCK_SIZE][j4][0] = pmvbw[0];
                  img->bw_mv[i4+BLOCK_SIZE][j4][1]=img->dbMV[i4+BLOCK_SIZE][j4][1] = pmvbw[1];
                  bwRefFrArr[j4][i4]=bw_rFrame;
                  if (img->mb_frame_field_flag)
                  {
                    if (img->current_mb_nr%2 == 0)
                    {
                      img->dbMV_top[i4+BLOCK_SIZE][j5][0]=img->bw_mv_top[i4+BLOCK_SIZE][j5][0]=pmvbw[0];
                      img->dbMV_top[i4+BLOCK_SIZE][j5][1]=img->bw_mv_top[i4+BLOCK_SIZE][j5][1]=pmvbw[1]/2;
                      img->bw_refFrArr_top[j5][i4]=2*bw_rFrame;
                    }
                    else
                    {
                      img->dbMV_bot[i4+BLOCK_SIZE][j5][0]=img->bw_mv_bot[i4+BLOCK_SIZE][j5][0]=pmvbw[0];
                      img->dbMV_bot[i4+BLOCK_SIZE][j5][1]=img->bw_mv_bot[i4+BLOCK_SIZE][j5][1]=pmvbw[1]/2;
                      img->bw_refFrArr_bot[j5][i4]=2*bw_rFrame;
                    }
                  }
                }               
              }
              else
              {                  
                img->bw_mv[i4+BLOCK_SIZE][j4][0]=img->dbMV[i4+BLOCK_SIZE][j4][0] = 0;
                img->bw_mv[i4+BLOCK_SIZE][j4][1]=img->dbMV[i4+BLOCK_SIZE][j4][1] = 0;
                bwRefFrArr[j6][i4]=-1;
                
                if (img->mb_frame_field_flag)
                {
                  if (img->current_mb_nr%2 == 0)
                  {
                    img->dbMV_top[i4+BLOCK_SIZE][j5][0]=img->bw_mv_top[i4+BLOCK_SIZE][j5][0]=0;
                    img->dbMV_top[i4+BLOCK_SIZE][j5][1]=img->bw_mv_top[i4+BLOCK_SIZE][j5][1]=0;
                    img->bw_refFrArr_top[j5][i4]=-1;
                  }
                  else
                  {
                    img->dbMV_bot[i4+BLOCK_SIZE][j5][0]=img->bw_mv_bot[i4+BLOCK_SIZE][j5][0]=0;
                    img->dbMV_bot[i4+BLOCK_SIZE][j5][1]=img->bw_mv_bot[i4+BLOCK_SIZE][j5][1]=0;
                    img->bw_refFrArr_bot[j5][i4]=-1;
                  }
                }
              }
              if (fw_rFrame ==15 && bw_rFrame ==15)
              {
                if (img->structure == TOP_FIELD)
                  fwRefFrArr[j6][i4] = 0;
                else
                  fwRefFrArr[j6][i4] = 0;
                bwRefFrArr[j6][i4] = (img->num_ref_idx_l1_active>1)?1:0;
                
                if (img->mb_frame_field_flag)
                {
                  if (img->current_mb_nr%2 == 0)
                  {
                    img->fw_refFrArr_top[j5][i4]=0;
                    img->bw_refFrArr_top[j5][i4]=2*bwRefFrArr[j6][i4];
                  }
                  else
                  {
                    img->fw_refFrArr_bot[j5][i4]=0; 
                    img->bw_refFrArr_bot[j5][i4]=2*bwRefFrArr[j6][i4];
                  }
                }
              }
              if (img->mb_frame_field_flag)
              {
                img->fw_mv_frm[i4+BLOCK_SIZE][j4][0]=img->fw_mv[i4+BLOCK_SIZE][j4][0];
                img->fw_mv_frm[i4+BLOCK_SIZE][j4][1]=img->fw_mv[i4+BLOCK_SIZE][j4][1];
                img->bw_mv_frm[i4+BLOCK_SIZE][j4][0]=img->bw_mv[i4+BLOCK_SIZE][j4][0];
                img->bw_mv_frm[i4+BLOCK_SIZE][j4][1]=img->bw_mv[i4+BLOCK_SIZE][j4][1];
              }
              
              // Revisit Alexis
              fw_refframe = (fwRefFrArr[j6][i4]!=-1) ? fwRefFrArr[j6][i4]:0;
              bw_refframe = (bwRefFrArr[j6][i4]!=-1) ? bwRefFrArr[j6][i4]:0;
              
              fw_ref_idx = (fwRefFrArr[j6][i4]!=-1) ? fwRefFrArr[j6][i4]:0;
              bw_ref_idx = (bwRefFrArr[j6][i4]!=-1) ? bwRefFrArr[j6][i4]:0;
              bw_ref_idx = img->disposable_flag ? bw_ref_idx : 1 - bw_ref_idx;
              
              if      (bwRefFrArr[j6][i4]==-1) direct_pdir = 0;
              else if (fwRefFrArr[j6][i4]==-1) direct_pdir = 1;
              else                             direct_pdir = 2;
              }
              else // Temporal Mode
              {
                if(refFrArr[j4][i4]==-1) // next P is intra mode
                {
                  for(hv=0; hv<2; hv++)
                  {
                    img->dfMV[i4+BLOCK_SIZE][j4][hv]=img->dbMV[i4+BLOCK_SIZE][j4][hv]=0;
                    img->fw_mv[i4+BLOCK_SIZE][j4][hv]=img->bw_mv[i4+BLOCK_SIZE][j4][hv]=0;
                    
                    if (img->mb_frame_field_flag)
                    {
                      if (img->current_mb_nr%2 == 0)
                      {
                        img->dfMV_top[i4+BLOCK_SIZE][j5][hv]=img->dbMV_top[i4+BLOCK_SIZE][j5][hv]=0;
                        img->fw_mv_top[i4+BLOCK_SIZE][j5][hv]=img->bw_mv_top[i4+BLOCK_SIZE][j5][hv]=0;
                      }
                      else
                      {
                        img->dfMV_bot[i4+BLOCK_SIZE][j5][hv]=img->dbMV_bot[i4+BLOCK_SIZE][j5][hv]=0;
                        img->fw_mv_bot[i4+BLOCK_SIZE][j5][hv]=img->bw_mv_bot[i4+BLOCK_SIZE][j5][hv]=0;
                      }
                    }
                  }
                  img->fw_refFrArr[j4][i4]=-1;
                  img->bw_refFrArr[j4][i4]=-1;
                  
                  if (img->mb_frame_field_flag)
                  {
                    if (img->current_mb_nr%2 == 0)
                    {
                      img->fw_refFrArr_top[j5][i4]=-1;
                      img->bw_refFrArr_top[j5][i4]=-1;
                    }
                    else
                    {
                      img->fw_refFrArr_bot[j5][i4]=-1;
                      img->bw_refFrArr_bot[j5][i4]=-1;
                    }
                  }
                  
                  if (img->structure == FRAME)
                  {
                    fw_refframe = 1;
                    fw_ref_idx = 0;
                  }
                  else
                  {
                    if (img->structure == TOP_FIELD)
                    {
                      fw_refframe = 2;    // used to be 3;
                      fw_ref_idx = 1;
                    }
                    else
                    {
                      fw_refframe = 1;    // used to be 2
                      fw_ref_idx = 0;
                    }
                  }
                }
                else // next P is skip or inter mode
                {                  
                  ref_idx = refFrArr[j4][i4];
                  fw_ref_idx = ref_idx;
                  frame_no_next_P =img->imgtr_next_P+((img->structure==BOTTOM_FIELD));
                  //frame_no_B = (mref==mref_fld) ? img->tr : 2*img->tr;
                  frame_no_B = (img->structure==TOP_FIELD || img->structure==BOTTOM_FIELD) ? img->tr_fld : 2*img->tr_frm;
                  delta_P = (img->imgtr_next_P - img->imgtr_last_P);
                  
                  if((img->structure==TOP_FIELD)) // top field
                  {
                    iTRp = delta_P*(ref_idx/2+1)-(ref_idx+1)%2;
                  }
                  else if((img->structure==BOTTOM_FIELD)) // bot field
                  {
                    iTRp = 1+delta_P*((ref_idx+1)/2)-ref_idx%2;
                  }
                  else  // frame
                  {
                    iTRp = (ref_idx+1)*delta_P;
                  }
                  
                  iTRb = iTRp - (frame_no_next_P - frame_no_B);
                  mv_scale = (iTRb * 256)/ iTRp;  //! Note that this could be precomputed at the frame/slice level. 
                  //! In such case, an array is needed for each different reference.
                  img->dfMV[i4+BLOCK_SIZE][j4][0]=(mv_scale * img->mv[i4+BLOCK_SIZE][j4][0] + 128 ) >> 8;
                  img->dfMV[i4+BLOCK_SIZE][j4][1]=(mv_scale * img->mv[i4+BLOCK_SIZE][j4][1] + 128 ) >> 8;
                  img->dbMV[i4+BLOCK_SIZE][j4][0]=((mv_scale-256) * img->mv[i4+BLOCK_SIZE][j4][0] + 128 ) >> 8;
                  img->dbMV[i4+BLOCK_SIZE][j4][1]=((mv_scale-256) * img->mv[i4+BLOCK_SIZE][j4][1] + 128 ) >> 8;
                  
                  if (img->mb_frame_field_flag)
                  {
                    if (img->current_mb_nr%2 == 0)
                    {
                      img->dfMV_top[i4+BLOCK_SIZE][j5][0]=(mv_scale * img->mv[i4+BLOCK_SIZE][j4][0] + 128 ) >> 8;
                      img->dfMV_top[i4+BLOCK_SIZE][j5][1]=(mv_scale/2 * img->mv[i4+BLOCK_SIZE][j4][1] + 128) >> 8;
                      img->dbMV_top[i4+BLOCK_SIZE][j5][0]=((mv_scale-256)*img->mv[i4+BLOCK_SIZE][j4][0] + 128 ) >> 8;
                    }
                    else
                    {
                      img->dfMV_bot[i4+BLOCK_SIZE][j5][0]=(mv_scale * img->mv[i4+BLOCK_SIZE][j4][0] + 128) >> 8;
                      img->dfMV_bot[i4+BLOCK_SIZE][j5][1]=(mv_scale/2 * img->mv[i4+BLOCK_SIZE][j4][1] + 128) >> 8;
                      img->dbMV_bot[i4+BLOCK_SIZE][j5][0]=((mv_scale-256)*img->mv[i4+BLOCK_SIZE][j4][0] + 128) >> 8;
                      img->dbMV_bot[i4+BLOCK_SIZE][j5][1]=((mv_scale/2-256)*img->mv[i4+BLOCK_SIZE][j4][1] + 128) >> 8;
                    }
                  }
                  
                  if (img->structure == TOP_FIELD)
                  {
                    fw_refframe = refFrArr[j4][i4]+2;
                    fw_ref_idx = refFrArr[j4][i4];
                  }
                  
                  else
                  {
                    fw_refframe = max(0,refFrArr[j4][i4]) + 1;  // DIRECT
                    fw_ref_idx = max(0,refFrArr[j4][i4]);
                  }
                  img->fw_mv[i4+BLOCK_SIZE][j4][0]=img->dfMV[i4+BLOCK_SIZE][j4][0];
                  img->fw_mv[i4+BLOCK_SIZE][j4][1]=img->dfMV[i4+BLOCK_SIZE][j4][1];
                  img->bw_mv[i4+BLOCK_SIZE][j4][0]=img->dbMV[i4+BLOCK_SIZE][j4][0];
                  img->bw_mv[i4+BLOCK_SIZE][j4][1]=img->dbMV[i4+BLOCK_SIZE][j4][1];
                  
                  if (img->mb_frame_field_flag)
                  {
                    if (img->current_mb_nr%2 == 0)
                    {
                      img->fw_mv_top[i4+BLOCK_SIZE][j5][0]=img->dfMV_top[i4+BLOCK_SIZE][j5][0]=img->fw_mv[i4+BLOCK_SIZE][j4][0];
                      img->fw_mv_top[i4+BLOCK_SIZE][j5][1]=img->dfMV_top[i4+BLOCK_SIZE][j5][1]=img->fw_mv[i4+BLOCK_SIZE][j4][1]/2;
                      img->bw_mv_top[i4+BLOCK_SIZE][j5][0]=img->dbMV_top[i4+BLOCK_SIZE][j5][0]=img->bw_mv[i4+BLOCK_SIZE][j4][0];
                      img->bw_mv_top[i4+BLOCK_SIZE][j5][1]=img->dbMV_top[i4+BLOCK_SIZE][j5][1]=img->bw_mv[i4+BLOCK_SIZE][j4][1]/2;
                      
                      img->fw_mv_frm[i4+BLOCK_SIZE][j4][0]=img->fw_mv[i4+BLOCK_SIZE][j4][0];
                      img->fw_mv_frm[i4+BLOCK_SIZE][j4][1]=img->fw_mv[i4+BLOCK_SIZE][j4][1];
                      img->bw_mv_frm[i4+BLOCK_SIZE][j4][0]=img->bw_mv[i4+BLOCK_SIZE][j4][0];
                      img->bw_mv_frm[i4+BLOCK_SIZE][j4][1]=img->bw_mv[i4+BLOCK_SIZE][j4][1];
                    }
                    else
                    {
                      img->fw_mv_bot[i4+BLOCK_SIZE][j5][0]=img->dfMV_bot[i4+BLOCK_SIZE][j5][0]=img->fw_mv[i4+BLOCK_SIZE][j4][0];
                      img->fw_mv_bot[i4+BLOCK_SIZE][j5][1]=img->dfMV_bot[i4+BLOCK_SIZE][j5][1]=img->fw_mv[i4+BLOCK_SIZE][j4][1]/2;
                      img->bw_mv_bot[i4+BLOCK_SIZE][j5][0]=img->dbMV_bot[i4+BLOCK_SIZE][j5][0]=img->bw_mv[i4+BLOCK_SIZE][j4][0];
                      img->bw_mv_bot[i4+BLOCK_SIZE][j5][1]=img->dbMV_bot[i4+BLOCK_SIZE][j5][1]=img->bw_mv[i4+BLOCK_SIZE][j4][1]/2;
                      
                      img->fw_mv_frm[i4+BLOCK_SIZE][j4][0]=img->fw_mv[i4+BLOCK_SIZE][j4][0];
                      img->fw_mv_frm[i4+BLOCK_SIZE][j4][1]=img->fw_mv[i4+BLOCK_SIZE][j4][1];
                      img->bw_mv_frm[i4+BLOCK_SIZE][j4][0]=img->bw_mv[i4+BLOCK_SIZE][j4][0];
                      img->bw_mv_frm[i4+BLOCK_SIZE][j4][1]=img->bw_mv[i4+BLOCK_SIZE][j4][1];
                    }
                  }
                  
                  if (img->structure == TOP_FIELD)
                    img->fw_refFrArr[j4][i4]=refFrArr[j4][i4] + 0;    // PLUS2, Krit, 7/05 (used to be + 1)
                  else
                    img->fw_refFrArr[j4][i4]=refFrArr[j4][i4];
                  img->bw_refFrArr[j4][i4]=0;
                  
                  if (img->mb_frame_field_flag)
                  {
                    if (img->current_mb_nr%2 == 0)
                    {
                      img->fw_refFrArr_top[j5][i4] = (refFrArr[j4][i4] == -1) ? -1:2*refFrArr[j4][i4];
                      img->bw_refFrArr_top[j5][i4] = (bw_refframe == -1) ? -1:2*bw_refframe;  // will always be 0
                    }
                    else
                    {
                      img->fw_refFrArr_bot[j5][i4] = (refFrArr[j4][i4] == -1) ? -1:2*refFrArr[j4][i4];
                      img->bw_refFrArr_bot[j5][i4] = (bw_refframe == -1) ? -1:2*bw_refframe;  // will always be 0
                    }
                  }
                }
              }
            }            
            if (mv_mode==0 && img->direct_type )
            {
              if (fwRefFrArr[j6][i4] >= 0)
              {
                vec1_x = i4*4*mv_mul + fw_mv_array[i4][j4][0];
                vec1_y = j4*4*mv_mul + fw_mv_array[i4][j4][1];
                get_block(fw_refframe, listX[0+list_offset], vec1_x, vec1_y, img, tmp_block);
              }
              
              if (bwRefFrArr[j6][i4] >= 0)
              {
                vec2_x = i4*4*mv_mul + bw_mv_array[i4][j4][0];
                vec2_y = j4*4*mv_mul + bw_mv_array[i4][j4][1];
                get_block(bw_refframe, listX[1+list_offset], vec2_x, vec2_y, img, tmp_blockbw);
              }              
            }
            else
            {
              vec1_x = i4*4*mv_mul + fw_mv_array[i4][j4][0];
              vec2_x = i4*4*mv_mul + bw_mv_array[i4][j4][0];

              if (!curr_mb_field)
              {
                vec1_y = j4*4*mv_mul + fw_mv_array[i4][j4][1];
                vec2_y = j4*4*mv_mul + bw_mv_array[i4][j4][1];
              }
              else
              {
                if (mb_nr%2 == 0)
                {
                  vec1_y = (img->block_y * 2 + joff) * mv_mul + fw_mv_array[i4][j4][1];
                  vec2_y = (img->block_y * 2 + joff) * mv_mul + bw_mv_array[i4][j4][1];
                }
                else
                {
                  vec1_y = ((img->block_y-4) * 2 + joff)* mv_mul + fw_mv_array[i4][j4][1];
                  vec2_y = ((img->block_y-4) * 2 + joff)* mv_mul + bw_mv_array[i4][j4][1];
                }
              }
              
              get_block(fw_refframe, listX[0+list_offset], vec1_x, vec1_y, img, tmp_block);
              get_block(bw_refframe, listX[1+list_offset], vec2_x, vec2_y, img, tmp_blockbw);
              
            }
            
            if (mv_mode==0 && img->direct_type && direct_pdir==0)
            {
              if (img->apply_weights)
              {
                for(ii=0;ii<BLOCK_SIZE;ii++)
                  for(jj=0;jj<BLOCK_SIZE;jj++)  
                    img->mpr[ii+ioff][jj+joff] = Clip1(((tmp_block[ii][jj] * img->wp_weight[0][fw_ref_idx][0]  + img->wp_round_luma)>>img->luma_log2_weight_denom) + img->wp_offset[0][fw_ref_idx][0]);
              }
              else
              {
                for(ii=0;ii<BLOCK_SIZE;ii++)
                  for(jj=0;jj<BLOCK_SIZE;jj++)  
                    img->mpr[ii+ioff][jj+joff] = tmp_block[ii][jj];
              }
            }
            else if (mv_mode==0 && img->direct_type && direct_pdir==1)
            {              
              if (img->apply_weights)
              {
                for(ii=0;ii<BLOCK_SIZE;ii++)
                  for(jj=0;jj<BLOCK_SIZE;jj++)  
                    img->mpr[ii+ioff][jj+joff] = Clip1(((tmp_blockbw[ii][jj] * img->wp_weight[1][bw_ref_idx][0] + img->wp_round_luma)>>img->luma_log2_weight_denom) + img->wp_offset[1][bw_ref_idx][0]);
              }
              else
              {
                for(ii=0;ii<BLOCK_SIZE;ii++)
                  for(jj=0;jj<BLOCK_SIZE;jj++)  
                    img->mpr[ii+ioff][jj+joff] = tmp_blockbw[ii][jj];
              }
            }
            else if(img->apply_weights)
            {
              int alpha_fw = img->wbp_weight[0][fw_ref_idx][bw_ref_idx][0];
              int alpha_bw = img->wbp_weight[1][fw_ref_idx][bw_ref_idx][0];
              
              for(ii=0;ii<BLOCK_SIZE;ii++)
                for(jj=0;jj<BLOCK_SIZE;jj++)  
                  img->mpr[ii+ioff][jj+joff] = (int)Clip1(((alpha_fw * tmp_block[ii][jj] + alpha_bw * tmp_blockbw[ii][jj]  + 2*img->wp_round_luma) >> (img->luma_log2_weight_denom+1)) + ((img->wp_offset[0][fw_ref_idx][0] + img->wp_offset[1][bw_ref_idx][0] + 1) >>1));
            }
            else
            {
              for(ii=0;ii<BLOCK_SIZE;ii++)
                for(jj=0;jj<BLOCK_SIZE;jj++)  
                  img->mpr[ii+ioff][jj+joff] = (tmp_block[ii][jj]+tmp_blockbw[ii][jj]+1)/2;
            }
          }
        }
        if (smb && mv_mode!=IBLOCK)
        {
          itrans_sp(img,ioff,joff,i,j);
        }
        else
        {
          itrans   (img,ioff,joff,i,j);      // use DCT transform and make 4x4 block m7 from prediction block mpr
        }
        
        for(ii=0;ii<BLOCK_SIZE;ii++)
        {
          for(jj=0;jj<BLOCK_SIZE;jj++)
          {
            dec_picture->imgY[j4*BLOCK_SIZE+jj][i4*BLOCK_SIZE+ii]=img->m7[ii][jj]; // contruct picture from 4x4 blocks
          }
        }
      }
  }
  
  // skipped MB processing for field MB  ***********************************
//  if (IS_COPY (currMB) && img->mb_frame_field_flag && img->current_mb_nr%2 ==0)      
//    decode_one_Copy_topMB(img,inp);

  // chroma decoding *******************************************************
  for(uv=0;uv<2;uv++)
  {
    intra_prediction = IS_INTRA (currMB);
    
    if (intra_prediction)
    {
      intrapred_chroma(img, uv);
    }
    
    for (j=4;j<6;j++)
    {
      joff=(j-4)*4;
      j4=img->pix_c_y+joff;
      for(i=0;i<2;i++)
      {
        ioff=i*4;
        i4=img->pix_c_x+ioff;
        
        mv_mode  = currMB->b8mode[2*(j-4)+i];
        pred_dir = currMB->b8pdir[2*(j-4)+i];
        assert (pred_dir<=2);

        if (!intra_prediction)
        {
          if (pred_dir != 2)
          {
            //--- FORWARD/BACKWARD PREDICTION ---
            mv_array = dec_picture->mv[LIST_0 + pred_dir];
            list = listX[0+list_offset+pred_dir];
            
            for(jj=0;jj<4;jj++)
            {
              jf=(j4+jj)/2;
              for(ii=0;ii<4;ii++)
              {
                if1=(i4+ii)/2;
                
                ref_idx   = dec_picture->ref_idx[LIST_0+pred_dir][if1][jf];
                
                i1=(img->pix_c_x+ii+ioff)*f1+mv_array[if1][jf][0];

                if (!curr_mb_field)
                  j1=(img->pix_c_y+jj+joff)*f1+mv_array[if1][jf][1];
                else
                {
                  if (mb_nr%2 == 0) 
                    j1=(img->pix_c_y)/2*f1 + (jj+joff)*f1+mv_array[if1][jf][1];
                  else
                    j1=(img->pix_c_y-8)/2*f1 + (jj+joff)*f1 +mv_array[if1][jf][1];
                }
                
                j1 += list[ref_idx]->chroma_vector_adjustment;
                
                ii0=max (0, min (i1/f1, img->width_cr-1));
                jj0=max (0, min (j1/f1, max_y_cr));
                ii1=max (0, min ((i1+f2)/f1, img->width_cr-1));
                jj1=max (0, min ((j1+f2)/f1, max_y_cr));
                
                if1=(i1 & f2);
                jf1=(j1 & f2);
                if0=f1-if1;
                jf0=f1-jf1;
                
                if (img->apply_weights)
                {
                  pred = (if0*jf0*list[ref_idx]->imgUV[uv][jj0][ii0]+
                          if1*jf0*list[ref_idx]->imgUV[uv][jj0][ii1]+
                          if0*jf1*list[ref_idx]->imgUV[uv][jj1][ii0]+
                          if1*jf1*list[ref_idx]->imgUV[uv][jj1][ii1]+f4)/f3;
                  img->mpr[ii+ioff][jj+joff] = Clip1(((img->wp_weight[pred_dir][ref_idx][uv+1] * pred  + img->wp_round_chroma)>>img->chroma_log2_weight_denom) + img->wp_offset[pred_dir][ref_idx][uv+1]);
                }
                else
                {
                  img->mpr[ii+ioff][jj+joff]=(if0*jf0*list[ref_idx]->imgUV[uv][jj0][ii0]+
                                              if1*jf0*list[ref_idx]->imgUV[uv][jj0][ii1]+
                                              if0*jf1*list[ref_idx]->imgUV[uv][jj1][ii0]+
                                              if1*jf1*list[ref_idx]->imgUV[uv][jj1][ii1]+f4)/f3;
                }
              }
            }
          }
          else
          {
            if (mv_mode != 0)
            {
              //===== BI-DIRECTIONAL PREDICTION =====
              fw_mv_array = dec_picture->mv[LIST_0];
              bw_mv_array = dec_picture->mv[LIST_1];
            }
            else
            {
              //===== DIRECT PREDICTION =====
              fw_mv_array = img->dfMV;
              bw_mv_array = img->dbMV;
            }
            
            for(jj=0;jj<4;jj++)
            {
              jf=(j4+jj)/2;
              for(ii=0;ii<4;ii++)
              {
                ifx=(i4+ii)/2;
                
                direct_pdir = 2;
                
                if (mv_mode != 0)
                {
                  fw_refframe = dec_picture->ref_idx[LIST_0][ifx][jf];
                  bw_refframe = dec_picture->ref_idx[LIST_1][ifx][jf];
                  
                  fw_ref_idx = fw_refframe;
                  bw_ref_idx = bw_refframe;
                  bw_ref_idx = img->disposable_flag ? bw_ref_idx : 1 - bw_ref_idx;
                }
                else
                {
                  if (!mv_mode && img->direct_type )
                  {
                    if (fwRefFrArr[2*(jf/2)][(ifx/2)*2]!=-1)
                    {
                      fw_refframe = fwRefFrArr[2*(jf/2)][(ifx/2)*2];
                      fw_ref_idx = fwRefFrArr[2*(jf/2)][(ifx/2)*2];
                    }
                    if (bwRefFrArr[2*(jf/2)][(ifx/2)*2]!=-1)
                    {
                      bw_refframe = bwRefFrArr[2*(jf/2)][(ifx/2)*2];
                      bw_ref_idx = bwRefFrArr[2*(jf/2)][(ifx/2)*2];
                      bw_ref_idx = img->disposable_flag ? bw_ref_idx : 1 - bw_ref_idx;
                    }
                    
                    if      (bwRefFrArr[2*(jf/2)][(ifx/2)*2]==-1) direct_pdir = 0;
                    else if (fwRefFrArr[2*(jf/2)][(ifx/2)*2]==-1) direct_pdir = 1;
                    else                                          direct_pdir = 2;
                  }                
                  else
                  {
                    bw_refframe = 0;
                    if (img->structure==TOP_FIELD)  
                    {
                      bw_refframe = 1;
                      bw_ref_idx = 1;
                      bw_ref_idx = img->disposable_flag ? bw_ref_idx : 1 - bw_ref_idx;
                    }
                    
                    if (img->structure == FRAME)
                    {
                      if(refFrArr[jf][ifx]==-1)  
                      {
                        fw_refframe = 1;
                        fw_ref_idx = 0; // revisit
                      }
                      else 
                      {
                        fw_refframe = 1+refFrArr[jf][ifx];
                        fw_ref_idx = refFrArr[jf][ifx]; //revisit
                      }
                    }
                    else
                    {
                      if (img->structure == TOP_FIELD)
                      {
                        if(refFrArr[jf][ifx]==-1)  
                        {
                          fw_refframe = 2;  // used to be 3; DIRECT
                          fw_ref_idx = 1; //revisit
                        }
                        else
                        {
                          fw_refframe = 2+refFrArr[jf][ifx];
                          fw_ref_idx = refFrArr[jf][ifx]; // revisit
                        }
                      }
                      else
                      {
                        if(refFrArr[jf][ifx]==-1)  
                        {
                          fw_refframe = 1;  // used to be 2; DIRECT
                          fw_ref_idx = 0; //revisit
                        }
                        else
                        {
                          fw_refframe = max(0,refFrArr[jf][ifx]) + 1;  // DIRECT
                          fw_ref_idx = refFrArr[jf][ifx]; //revisit
                        }
                        // used to have 1+refFrArr[jf][ifx]; DIRECT
                      }
                    }
                  }
                }
                

                if (mv_mode==0 && img->direct_type )
                {
                  if (direct_pdir == 0 || direct_pdir == 2)
                  {
                    i1=(img->pix_c_x+ii+ioff)*f1+fw_mv_array[ifx][jf][0];
                    j1=(img->pix_c_y+jj+joff)*f1+fw_mv_array[ifx][jf][1];
                    j1 += listX[0+list_offset][fw_refframe]->chroma_vector_adjustment;
                    
                    ii0=max (0, min (i1/f1, img->width_cr-1));
                    jj0=max (0, min (j1/f1, img->height_cr-1));
                    ii1=max (0, min ((i1+f2)/f1, img->width_cr-1));
                    jj1=max (0, min ((j1+f2)/f1, img->height_cr-1));
                    
                    if1=(i1 & f2);
                    jf1=(j1 & f2);
                    if0=f1-if1;
                    jf0=f1-jf1;
                    
                    fw_pred=(if0*jf0*listX[0+list_offset][fw_refframe]->imgUV[uv][jj0][ii0]+
                             if1*jf0*listX[0+list_offset][fw_refframe]->imgUV[uv][jj0][ii1]+
                             if0*jf1*listX[0+list_offset][fw_refframe]->imgUV[uv][jj1][ii0]+
                             if1*jf1*listX[0+list_offset][fw_refframe]->imgUV[uv][jj1][ii1]+f4)/f3;
                  }
                  if (direct_pdir == 1 || direct_pdir == 2)
                  {
                    i1=(img->pix_c_x+ii+ioff)*f1+bw_mv_array[ifx][jf][0];
                    j1=(img->pix_c_y+jj+joff)*f1+bw_mv_array[ifx][jf][1];
                    j1 += listX[1+list_offset][bw_refframe]->chroma_vector_adjustment;
                    
                    ii0=max (0, min (i1/f1, img->width_cr-1));
                    jj0=max (0, min (j1/f1, img->height_cr-1));
                    ii1=max (0, min ((i1+f2)/f1, img->width_cr-1));
                    jj1=max (0, min ((j1+f2)/f1, img->height_cr-1));
                    
                    if1=(i1 & f2);
                    jf1=(j1 & f2);
                    if0=f1-if1;
                    jf0=f1-jf1;
                
                    bw_pred=(if0*jf0*listX[1+list_offset][bw_refframe]->imgUV[uv][jj0][ii0]+
                             if1*jf0*listX[1+list_offset][bw_refframe]->imgUV[uv][jj0][ii1]+
                             if0*jf1*listX[1+list_offset][bw_refframe]->imgUV[uv][jj1][ii0]+
                             if1*jf1*listX[1+list_offset][bw_refframe]->imgUV[uv][jj1][ii1]+f4)/f3;
                  }

                }
                else
                {
                  i1=(img->pix_c_x+ii+ioff)*f1+fw_mv_array[ifx][jf][0];

                  if (!curr_mb_field)
                  {
                    j1=(img->pix_c_y+jj+joff)*f1+fw_mv_array[ifx][jf][1];
                  }
                  else
                  {
                    if (mb_nr%2 == 0) 
                      j1=(img->pix_c_y)/2*f1 + (jj+joff)*f1+fw_mv_array[ifx][jf][1];
                    else
                      j1=(img->pix_c_y-8)/2*f1 + (jj+joff)*f1 +fw_mv_array[ifx][jf][1];
                  }

                  j1 += listX[0+list_offset][fw_refframe]->chroma_vector_adjustment;
                  
                  ii0=max (0, min (i1/f1, img->width_cr-1));
                  jj0=max (0, min (j1/f1, max_y_cr));
                  ii1=max (0, min ((i1+f2)/f1, img->width_cr-1));
                  jj1=max (0, min ((j1+f2)/f1, max_y_cr));
                  
                  if1=(i1 & f2);
                  jf1=(j1 & f2);
                  if0=f1-if1;
                  jf0=f1-jf1;
                  
                  fw_pred=(if0*jf0*listX[0+list_offset][fw_refframe]->imgUV[uv][jj0][ii0]+
                           if1*jf0*listX[0+list_offset][fw_refframe]->imgUV[uv][jj0][ii1]+
                           if0*jf1*listX[0+list_offset][fw_refframe]->imgUV[uv][jj1][ii0]+
                           if1*jf1*listX[0+list_offset][fw_refframe]->imgUV[uv][jj1][ii1]+f4)/f3;
                  
                  i1=(img->pix_c_x+ii+ioff)*f1+bw_mv_array[ifx][jf][0];
                  
                  if (!curr_mb_field)
                  {
                    j1=(img->pix_c_y+jj+joff)*f1+bw_mv_array[ifx][jf][1];
                  }
                  else
                  {
                    if (mb_nr%2 == 0) 
                      j1=(img->pix_c_y)/2*f1 + (jj+joff)*f1+bw_mv_array[ifx][jf][1];
                    else
                      j1=(img->pix_c_y-8)/2*f1 + (jj+joff)*f1 +bw_mv_array[ifx][jf][1];
                  }

                  j1 += listX[1+list_offset][bw_refframe]->chroma_vector_adjustment;

                  ii0=max (0, min (i1/f1, img->width_cr-1));
                  jj0=max (0, min (j1/f1, max_y_cr));
                  ii1=max (0, min ((i1+f2)/f1, img->width_cr-1));
                  jj1=max (0, min ((j1+f2)/f1, max_y_cr));
                  
                  if1=(i1 & f2);
                  jf1=(j1 & f2);
                  if0=f1-if1;
                  jf0=f1-jf1;
                  
                  bw_pred=(if0*jf0*listX[1+list_offset][bw_refframe]->imgUV[uv][jj0][ii0]+
                           if1*jf0*listX[1+list_offset][bw_refframe]->imgUV[uv][jj0][ii1]+
                           if0*jf1*listX[1+list_offset][bw_refframe]->imgUV[uv][jj1][ii0]+
                           if1*jf1*listX[1+list_offset][bw_refframe]->imgUV[uv][jj1][ii1]+f4)/f3;
                  
                }
                
                if (img->apply_weights)
                {
                  if (img->direct_type && direct_pdir==1)
                  {
                    img->mpr[ii+ioff][jj+joff]= Clip1(((img->wp_weight[1][bw_ref_idx][uv+1] * bw_pred  + img->wp_round_chroma)>>img->chroma_log2_weight_denom) + img->wp_offset[1][bw_ref_idx][uv+1]);   //<! Replaced with integer only operations
                  }
                  else if (img->direct_type && direct_pdir==0)
                  {
                    img->mpr[ii+ioff][jj+joff]=Clip1(((img->wp_weight[0][fw_ref_idx][uv+1] * fw_pred + img->wp_round_chroma)>>img->chroma_log2_weight_denom) + img->wp_offset[0][fw_ref_idx][uv+1]);   //<! Replaced with integer only operations
                  }
                  else
                  {
                    
                    int alpha_fw = img->wbp_weight[0][fw_ref_idx][bw_ref_idx][uv+1];
                    int alpha_bw = img->wbp_weight[1][fw_ref_idx][bw_ref_idx][uv+1];
                    img->mpr[ii+ioff][jj+joff]= Clip1(((alpha_fw * fw_pred + alpha_bw * bw_pred  + 2*img->wp_round_chroma) >> (img->chroma_log2_weight_denom + 1))+ ((img->wp_offset[0][fw_ref_idx][uv+1] + img->wp_offset[1][bw_ref_idx][uv+1] + 1)>>1) );
                  }
                }
                else
                {
                  if (img->direct_type && direct_pdir==1)
                  {
                    img->mpr[ii+ioff][jj+joff]=bw_pred;
                  }
                  else if (img->direct_type && direct_pdir==0)
                  {
                    img->mpr[ii+ioff][jj+joff]=fw_pred;
                  } 
                  else
                  {
                    img->mpr[ii+ioff][jj+joff]=(fw_pred + bw_pred + 1 )/2;
                  }
                }
              }
            }
          }
        }
        
        if (!smb)
        {
          itrans(img,ioff,joff,2*uv+i,j);
          for(ii=0;ii<4;ii++)
            for(jj=0;jj<4;jj++)
            {
              dec_picture->imgUV[uv][j4+jj][i4+ii]=img->m7[ii][jj];
            }
        }
      }
    }
    
    if(smb)
    {
      itrans_sp_chroma(img,2*uv);
      for (j=4;j<6;j++)
      {
        joff=(j-4)*4;
        j4=img->pix_c_y+joff;
        for(i=0;i<2;i++)
        {
          ioff=i*4;
          i4=img->pix_c_x+ioff;
          itrans(img,ioff,joff,2*uv+i,j);
          
          for(ii=0;ii<4;ii++)
            for(jj=0;jj<4;jj++)
            {
              dec_picture->imgUV[uv][j4+jj][i4+ii]=img->m7[ii][jj];
            }
        }
      }
    }
  }

  return 0;
}


